import logging
from xml.etree import cElementTree as eTree

from bustoken import TK_ELECTRONIC_PAYMENT_ADD_PAYMENT
from model.customexception import ElectronicPaymentException
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
    MBEasyContext,
)
from sysactions import show_messagebox
from systools import sys_log_exception
from typing import (
    Optional,
    List,
)

from actions.eft import (
    get_electronic_payment_component_name,
    get_tender_id
)

logger = logging.getLogger("PosActions")


def request_eft(
    pos_id,                     # type: str
    mb_context,                 # type: MBEasyContext
    model,                      # type: eTree.XML
    electronic_type,            # type: str
    amount,                     # type: float
    _eft_events,                # type: List
    is_table_service=False,     # type: bool
):
    # type: (...) -> Optional[str]

    from poslistener import eft_status_finish

    order = model.find("CurrentOrder/Order")
    totaled_state = order.findall("StateHistory/State[@state='TOTALED']")[-1]
    last_totaled = totaled_state.get("timestamp").replace("-", "").replace(":", "")
    order_id = order.get("orderId")
    all_tenders = order.find("TenderHistory").findall("Tender")
    tender_id = get_tender_id(all_tenders, is_table_service)
    fiscal_date = last_totaled[:8]
    fiscal_time = last_totaled[9:15]
    operator = model.find("Operator").get("id")
    session_id = model.find("Operator").get("sessionId")
    eft_name = get_electronic_payment_component_name(pos_id=pos_id)
    if eft_name is None:
        raise Exception("None eft found")

    try:
        component_name = eft_name
        evt_data = [eft_name, None, pos_id, order_id]
        _eft_events.append(evt_data)

        data = "%s;%s;%s;%s;%s;%s;%s;%s;%s" % (
            str(pos_id),
            str(order_id),
            str(session_id),
            str(operator),
            str(electronic_type),
            str(amount),
            fiscal_date,
            fiscal_time,
            tender_id,
        )

        log_msg = "Performing electronic payment. {}#{}. TenderId/Amount: {}/{}"
        logger.info(log_msg.format(pos_id, order_id, tender_id, amount))

        response = mb_context.MB_EasySendMessage(
            dest_name=component_name,
            token=TK_ELECTRONIC_PAYMENT_ADD_PAYMENT,
            format=FM_PARAM,
            data=data
        )
        if response.token != TK_SYS_ACK:
            raise ElectronicPaymentException(response.data)

        return response.data
    except Exception as ex: # noqa
        show_messagebox(pos_id, "$ERROR_PROCESSING_TEF|{}".format(ex.message), "$EFT", asynch=True, buttons="$OK")
        error_msg = "Error processing TEF. {}#{}".format(pos_id, order_id)
        logger.exception(error_msg)
        sys_log_exception(error_msg)
        raise
    finally:
        eft_status_finish([pos_id])



# banknote.py - functions for handling Euro banknote serial numbers
#
# Copyright (C) 2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Euro banknote serial numbers.

The serial number consists of one letter and 11 digits, or two letters and 10
digits for the new series banknotes.

More information:

* https://en.wikipedia.org/wiki/Euro_banknotes#Serial_number

>>> validate('P36007033744')
'P36007033744'
>>> validate('P36007033743')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').upper().strip()


def checksum(number):
    """Calculate the checksum over the number."""
    # replace letters by their ASCII number
    return sum(int(x) if isdigits(x) else ord(x) for x in number) % 9


def validate(number):
    """Check if the number is a valid banknote serial number."""
    number = compact(number)
    if not number[:2].isalnum() or not isdigits(number[2:]):
        raise InvalidFormat()
    if len(number) != 12:
        raise InvalidLength()
    if number[0] not in 'BCDEFGHJLMNPRSTUVWXYZ':
        raise InvalidComponent()
    if checksum(number) != 0:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid banknote serial number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# peid.py - library for Liechtenstein Personenidentifikationsnummer
#
# Copyright (C) 2020 Matthias Schmid
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""PEID (Liechtenstein tax code for individuals and entities).

The Personenidentifikationsnummer (PEID) is an numeric code up to 12 digits
used to identify entities and individuals residing in Liechtenstein.

More information:

* https://www.oera.li/
* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/LIECHTENSTEIN%20TIN.pdf

>>> compact('00001234567')
'1234567'
>>> validate('1234567')  # personal number or entity number
'1234567'
>>> validate('00001234567')
'1234567'
>>> validate('00001234568913454545')
Traceback (most recent call last):
    ...
InvalidLength: The number has an invalid length.
"""  # noqa: E501

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' .').strip().lstrip('0')


def validate(number):
    """Check if the given fiscal code is valid."""
    number = compact(number)
    if len(number) < 4 or len(number) > 12:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()

    return number


def is_valid(number):
    """Check if the given fiscal code is valid."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gc
import sys
import time
import threading

from abc import ABCMeta, abstractmethod

import greenlet
from greenlet import greenlet as RawGreenlet
from . import TestCase
from .leakcheck import fails_leakcheck


# We manually manage locks in many tests
# pylint:disable=consider-using-with
# pylint:disable=too-many-public-methods

class SomeError(Exception):
    pass


def fmain(seen):
    try:
        greenlet.getcurrent().parent.switch()
    except:
        seen.append(sys.exc_info()[0])
        raise
    raise SomeError


def send_exception(g, exc):
    # note: send_exception(g, exc)  can be now done with  g.throw(exc).
    # the purpose of this test is to explicitly check the propagation rules.
    def crasher(exc):
        raise exc
    g1 = RawGreenlet(crasher, parent=g)
    g1.switch(exc)


class TestGreenlet(TestCase):

    def _do_simple_test(self):
        lst = []

        def f():
            lst.append(1)
            greenlet.getcurrent().parent.switch()
            lst.append(3)
        g = RawGreenlet(f)
        lst.append(0)
        g.switch()
        lst.append(2)
        g.switch()
        lst.append(4)
        self.assertEqual(lst, list(range(5)))

    def test_simple(self):
        self._do_simple_test()

    def test_switch_no_run_raises_AttributeError(self):
        g = RawGreenlet()
        with self.assertRaises(AttributeError) as exc:
            g.switch()

        self.assertIn("run", str(exc.exception))

    def test_throw_no_run_raises_AttributeError(self):
        g = RawGreenlet()
        with self.assertRaises(AttributeError) as exc:
            g.throw(SomeError)

        self.assertIn("run", str(exc.exception))

    def test_parent_equals_None(self):
        g = RawGreenlet(parent=None)
        self.assertIsNotNone(g)
        self.assertIs(g.parent, greenlet.getcurrent())

    def test_run_equals_None(self):
        g = RawGreenlet(run=None)
        self.assertIsNotNone(g)
        self.assertIsNone(g.run)

    def test_two_children(self):
        lst = []

        def f():
            lst.append(1)
            greenlet.getcurrent().parent.switch()
            lst.extend([1, 1])
        g = RawGreenlet(f)
        h = RawGreenlet(f)
        g.switch()
        self.assertEqual(len(lst), 1)
        h.switch()
        self.assertEqual(len(lst), 2)
        h.switch()
        self.assertEqual(len(lst), 4)
        self.assertEqual(h.dead, True)
        g.switch()
        self.assertEqual(len(lst), 6)
        self.assertEqual(g.dead, True)

    def test_two_recursive_children(self):
        lst = []

        def f():
            lst.append('b')
            greenlet.getcurrent().parent.switch()

        def g():
            lst.append('a')
            g = RawGreenlet(f)
            g.switch()
            lst.append('c')

        g = RawGreenlet(g)
        self.assertEqual(sys.getrefcount(g), 2)
        g.switch()
        self.assertEqual(lst, ['a', 'b', 'c'])
        # Just the one in this frame, plus the one on the stack we pass to the function
        self.assertEqual(sys.getrefcount(g), 2)

    def test_threads(self):
        success = []

        def f():
            self._do_simple_test()
            success.append(True)
        ths = [threading.Thread(target=f) for i in range(10)]
        for th in ths:
            th.start()
        for th in ths:
            th.join(10)
        self.assertEqual(len(success), len(ths))

    def test_exception(self):
        seen = []
        g1 = RawGreenlet(fmain)
        g2 = RawGreenlet(fmain)
        g1.switch(seen)
        g2.switch(seen)
        g2.parent = g1

        self.assertEqual(seen, [])
        #with self.assertRaises(SomeError):
        #    p("***Switching back")
        #    g2.switch()
        # Creating this as a bound method can reveal bugs that
        # are hidden on newer versions of Python that avoid creating
        # bound methods for direct expressions; IOW, don't use the `with`
        # form!
        self.assertRaises(SomeError, g2.switch)
        self.assertEqual(seen, [SomeError])

        value = g2.switch()
        self.assertEqual(value, ())
        self.assertEqual(seen, [SomeError])

        value = g2.switch(25)
        self.assertEqual(value, 25)
        self.assertEqual(seen, [SomeError])


    def test_send_exception(self):
        seen = []
        g1 = RawGreenlet(fmain)
        g1.switch(seen)
        self.assertRaises(KeyError, send_exception, g1, KeyError)
        self.assertEqual(seen, [KeyError])

    def test_dealloc(self):
        seen = []
        g1 = RawGreenlet(fmain)
        g2 = RawGreenlet(fmain)
        g1.switch(seen)
        g2.switch(seen)
        self.assertEqual(seen, [])
        del g1
        gc.collect()
        self.assertEqual(seen, [greenlet.GreenletExit])
        del g2
        gc.collect()
        self.assertEqual(seen, [greenlet.GreenletExit, greenlet.GreenletExit])

    def test_dealloc_catches_GreenletExit_throws_other(self):
        def run():
            try:
                greenlet.getcurrent().parent.switch()
            except greenlet.GreenletExit:
                raise SomeError from None

        g = RawGreenlet(run)
        g.switch()
        # Destroying the only reference to the greenlet causes it
        # to get GreenletExit; when it in turn raises, even though we're the parent
        # we don't get the exception, it just gets printed.
        # When we run on 3.8 only, we can use sys.unraisablehook
        oldstderr = sys.stderr
        try:
            from cStringIO import StringIO
        except ImportError:
            from io import StringIO
        stderr = sys.stderr = StringIO()
        try:
            del g
        finally:
            sys.stderr = oldstderr

        v = stderr.getvalue()
        self.assertIn("Exception", v)
        self.assertIn('ignored', v)
        self.assertIn("SomeError", v)


    def test_dealloc_other_thread(self):
        seen = []
        someref = []

        bg_glet_created_running_and_no_longer_ref_in_bg = threading.Event()
        fg_ref_released = threading.Event()
        bg_should_be_clear = threading.Event()
        ok_to_exit_bg_thread = threading.Event()

        def f():
            g1 = RawGreenlet(fmain)
            g1.switch(seen)
            someref.append(g1)
            del g1
            gc.collect()

            bg_glet_created_running_and_no_longer_ref_in_bg.set()
            fg_ref_released.wait(3)

            RawGreenlet()   # trigger release
            bg_should_be_clear.set()
            ok_to_exit_bg_thread.wait(3)
            RawGreenlet() # One more time

        t = threading.Thread(target=f)
        t.start()
        bg_glet_created_running_and_no_longer_ref_in_bg.wait(10)

        self.assertEqual(seen, [])
        self.assertEqual(len(someref), 1)
        del someref[:]
        gc.collect()
        # g1 is not released immediately because it's from another thread
        self.assertEqual(seen, [])
        fg_ref_released.set()
        bg_should_be_clear.wait(3)
        try:
            self.assertEqual(seen, [greenlet.GreenletExit])
        finally:
            ok_to_exit_bg_thread.set()
            t.join(10)
            del seen[:]
            del someref[:]

    def test_frame(self):
        def f1():
            f = sys._getframe(0) # pylint:disable=protected-access
            self.assertEqual(f.f_back, None)
            greenlet.getcurrent().parent.switch(f)
            return "meaning of life"
        g = RawGreenlet(f1)
        frame = g.switch()
        self.assertTrue(frame is g.gr_frame)
        self.assertTrue(g)

        from_g = g.switch()
        self.assertFalse(g)
        self.assertEqual(from_g, 'meaning of life')
        self.assertEqual(g.gr_frame, None)

    def test_thread_bug(self):
        def runner(x):
            g = RawGreenlet(lambda: time.sleep(x))
            g.switch()
        t1 = threading.Thread(target=runner, args=(0.2,))
        t2 = threading.Thread(target=runner, args=(0.3,))
        t1.start()
        t2.start()
        t1.join(10)
        t2.join(10)

    def test_switch_kwargs(self):
        def run(a, b):
            self.assertEqual(a, 4)
            self.assertEqual(b, 2)
            return 42
        x = RawGreenlet(run).switch(a=4, b=2)
        self.assertEqual(x, 42)

    def test_switch_kwargs_to_parent(self):
        def run(x):
            greenlet.getcurrent().parent.switch(x=x)
            greenlet.getcurrent().parent.switch(2, x=3)
            return x, x ** 2
        g = RawGreenlet(run)
        self.assertEqual({'x': 3}, g.switch(3))
        self.assertEqual(((2,), {'x': 3}), g.switch())
        self.assertEqual((3, 9), g.switch())

    def test_switch_to_another_thread(self):
        data = {}
        created_event = threading.Event()
        done_event = threading.Event()

        def run():
            data['g'] = RawGreenlet(lambda: None)
            created_event.set()
            done_event.wait(10)
        thread = threading.Thread(target=run)
        thread.start()
        created_event.wait(10)
        with self.assertRaises(greenlet.error):
            data['g'].switch()
        done_event.set()
        thread.join(10)
        # XXX: Should handle this automatically
        data.clear()

    def test_exc_state(self):
        def f():
            try:
                raise ValueError('fun')
            except: # pylint:disable=bare-except
                exc_info = sys.exc_info()
                RawGreenlet(h).switch()
                self.assertEqual(exc_info, sys.exc_info())

        def h():
            self.assertEqual(sys.exc_info(), (None, None, None))

        RawGreenlet(f).switch()

    def test_instance_dict(self):
        def f():
            greenlet.getcurrent().test = 42
        def deldict(g):
            del g.__dict__
        def setdict(g, value):
            g.__dict__ = value
        g = RawGreenlet(f)
        self.assertEqual(g.__dict__, {})
        g.switch()
        self.assertEqual(g.test, 42)
        self.assertEqual(g.__dict__, {'test': 42})
        g.__dict__ = g.__dict__
        self.assertEqual(g.__dict__, {'test': 42})
        self.assertRaises(TypeError, deldict, g)
        self.assertRaises(TypeError, setdict, g, 42)

    def test_running_greenlet_has_no_run(self):
        has_run = []
        def func():
            has_run.append(
                hasattr(greenlet.getcurrent(), 'run')
            )

        g = RawGreenlet(func)
        g.switch()
        self.assertEqual(has_run, [False])

    def test_deepcopy(self):
        import copy
        self.assertRaises(TypeError, copy.copy, RawGreenlet())
        self.assertRaises(TypeError, copy.deepcopy, RawGreenlet())

    def test_parent_restored_on_kill(self):
        hub = RawGreenlet(lambda: None)
        main = greenlet.getcurrent()
        result = []
        def worker():
            try:
                # Wait to be killed by going back to the test.
                main.switch()
            except greenlet.GreenletExit:
                # Resurrect and switch to parent
                result.append(greenlet.getcurrent().parent)
                result.append(greenlet.getcurrent())
                hub.switch()
        g = RawGreenlet(worker, parent=hub)
        g.switch()
        # delete the only reference, thereby raising GreenletExit
        del g
        self.assertTrue(result)
        self.assertIs(result[0], main)
        self.assertIs(result[1].parent, hub)
        # Delete them, thereby breaking the cycle between the greenlet
        # and the frame, which otherwise would never be collectable
        # XXX: We should be able to automatically fix this.
        del result[:]
        hub = None
        main = None

    def test_parent_return_failure(self):
        # No run causes AttributeError on switch
        g1 = RawGreenlet()
        # Greenlet that implicitly switches to parent
        g2 = RawGreenlet(lambda: None, parent=g1)
        # AttributeError should propagate to us, no fatal errors
        with self.assertRaises(AttributeError):
            g2.switch()

    def test_throw_exception_not_lost(self):
        class mygreenlet(RawGreenlet):
            def __getattribute__(self, name):
                try:
                    raise Exception()
                except: # pylint:disable=bare-except
                    pass
                return RawGreenlet.__getattribute__(self, name)
        g = mygreenlet(lambda: None)
        self.assertRaises(SomeError, g.throw, SomeError())

    @fails_leakcheck
    def _do_test_throw_to_dead_thread_doesnt_crash(self, wait_for_cleanup=False):
        result = []
        def worker():
            greenlet.getcurrent().parent.switch()

        def creator():
            g = RawGreenlet(worker)
            g.switch()
            result.append(g)
            if wait_for_cleanup:
                # Let this greenlet eventually be cleaned up.
                g.switch()
                greenlet.getcurrent()
        t = threading.Thread(target=creator)
        t.start()
        t.join(10)
        del t
        # But, depending on the operating system, the thread
        # deallocator may not actually have run yet! So we can't be
        # sure about the error message unless we wait.
        if wait_for_cleanup:
            self.wait_for_pending_cleanups()
        with self.assertRaises(greenlet.error) as exc:
            result[0].throw(SomeError)

        if not wait_for_cleanup:
            self.assertIn(
                str(exc.exception), [
                    "cannot switch to a different thread (which happens to have exited)",
                    "cannot switch to a different thread"
                ]
            )
        else:
            self.assertEqual(
                str(exc.exception),
                "cannot switch to a different thread (which happens to have exited)",
            )

        if hasattr(result[0].gr_frame, 'clear'):
            # The frame is actually executing (it thinks), we can't clear it.
            with self.assertRaises(RuntimeError):
                result[0].gr_frame.clear()
        # Unfortunately, this doesn't actually clear the references, they're in the
        # fast local array.
        if not wait_for_cleanup:
            result[0].gr_frame.f_locals.clear()
        else:
            self.assertIsNone(result[0].gr_frame)

        del creator
        worker = None
        del result[:]
        # XXX: we ought to be able to automatically fix this.
        # See issue 252
        self.expect_greenlet_leak = True # direct us not to wait for it to go away

    @fails_leakcheck
    def test_throw_to_dead_thread_doesnt_crash(self):
        self._do_test_throw_to_dead_thread_doesnt_crash()

    def test_throw_to_dead_thread_doesnt_crash_wait(self):
        self._do_test_throw_to_dead_thread_doesnt_crash(True)

    @fails_leakcheck
    def test_recursive_startup(self):
        class convoluted(RawGreenlet):
            def __init__(self):
                RawGreenlet.__init__(self)
                self.count = 0
            def __getattribute__(self, name):
                if name == 'run' and self.count == 0:
                    self.count = 1
                    self.switch(43)
                return RawGreenlet.__getattribute__(self, name)
            def run(self, value):
                while True:
                    self.parent.switch(value)
        g = convoluted()
        self.assertEqual(g.switch(42), 43)
        # Exits the running greenlet, otherwise it leaks
        # XXX: We should be able to automatically fix this
        #g.throw(greenlet.GreenletExit)
        #del g
        self.expect_greenlet_leak = True

    def test_threaded_updatecurrent(self):
        # released when main thread should execute
        lock1 = threading.Lock()
        lock1.acquire()
        # released when another thread should execute
        lock2 = threading.Lock()
        lock2.acquire()
        class finalized(object):
            def __del__(self):
                # happens while in green_updatecurrent() in main greenlet
                # should be very careful not to accidentally call it again
                # at the same time we must make sure another thread executes
                lock2.release()
                lock1.acquire()
                # now ts_current belongs to another thread
        def deallocator():
            greenlet.getcurrent().parent.switch()
        def fthread():
            lock2.acquire()
            greenlet.getcurrent()
            del g[0]
            lock1.release()
            lock2.acquire()
            greenlet.getcurrent()
            lock1.release()
        main = greenlet.getcurrent()
        g = [RawGreenlet(deallocator)]
        g[0].bomb = finalized()
        g[0].switch()
        t = threading.Thread(target=fthread)
        t.start()
        # let another thread grab ts_current and deallocate g[0]
        lock2.release()
        lock1.acquire()
        # this is the corner stone
        # getcurrent() will notice that ts_current belongs to another thread
        # and start the update process, which would notice that g[0] should
        # be deallocated, and that will execute an object's finalizer. Now,
        # that object will let another thread run so it can grab ts_current
        # again, which would likely crash the interpreter if there's no
        # check for this case at the end of green_updatecurrent(). This test
        # passes if getcurrent() returns correct result, but it's likely
        # to randomly crash if it's not anyway.
        self.assertEqual(greenlet.getcurrent(), main)
        # wait for another thread to complete, just in case
        t.join(10)

    def test_dealloc_switch_args_not_lost(self):
        seen = []
        def worker():
            # wait for the value
            value = greenlet.getcurrent().parent.switch()
            # delete all references to ourself
            del worker[0]
            initiator.parent = greenlet.getcurrent().parent
            # switch to main with the value, but because
            # ts_current is the last reference to us we
            # return here immediately, where we resurrect ourself.
            try:
                greenlet.getcurrent().parent.switch(value)
            finally:
                seen.append(greenlet.getcurrent())
        def initiator():
            return 42 # implicitly falls thru to parent

        worker = [RawGreenlet(worker)]

        worker[0].switch() # prime worker
        initiator = RawGreenlet(initiator, worker[0])
        value = initiator.switch()
        self.assertTrue(seen)
        self.assertEqual(value, 42)

    def test_tuple_subclass(self):
        # The point of this test is to see what happens when a custom
        # tuple subclass is used as an object passed directly to the C
        # function ``green_switch``; part of ``green_switch`` checks
        # the ``len()`` of the ``args`` tuple, and that can call back
        # into Python. Here, when it calls back into Python, we
        # recursively enter ``green_switch`` again.

        # This test is really only relevant on Python 2. The builtin
        # `apply` function directly passes the given args tuple object
        # to the underlying function, whereas the Python 3 version
        # unpacks and repacks into an actual tuple. This could still
        # happen using the C API on Python 3 though. We should write a
        # builtin version of apply() ourself.
        def _apply(func, a, k):
            func(*a, **k)

        class mytuple(tuple):
            def __len__(self):
                greenlet.getcurrent().switch()
                return tuple.__len__(self)
        args = mytuple()
        kwargs = dict(a=42)
        def switchapply():
            _apply(greenlet.getcurrent().parent.switch, args, kwargs)
        g = RawGreenlet(switchapply)
        self.assertEqual(g.switch(), kwargs)

    def test_abstract_subclasses(self):
        AbstractSubclass = ABCMeta(
            'AbstractSubclass',
            (RawGreenlet,),
            {'run': abstractmethod(lambda self: None)})

        class BadSubclass(AbstractSubclass):
            pass

        class GoodSubclass(AbstractSubclass):
            def run(self):
                pass

        GoodSubclass() # should not raise
        self.assertRaises(TypeError, BadSubclass)

    def test_implicit_parent_with_threads(self):
        if not gc.isenabled():
            return # cannot test with disabled gc
        N = gc.get_threshold()[0]
        if N < 50:
            return # cannot test with such a small N
        def attempt():
            lock1 = threading.Lock()
            lock1.acquire()
            lock2 = threading.Lock()
            lock2.acquire()
            recycled = [False]
            def another_thread():
                lock1.acquire() # wait for gc
                greenlet.getcurrent() # update ts_current
                lock2.release() # release gc
            t = threading.Thread(target=another_thread)
            t.start()
            class gc_callback(object):
                def __del__(self):
                    lock1.release()
                    lock2.acquire()
                    recycled[0] = True
            class garbage(object):
                def __init__(self):
                    self.cycle = self
                    self.callback = gc_callback()
            l = []
            x = range(N*2)
            current = greenlet.getcurrent()
            g = garbage()
            for _ in x:
                g = None # lose reference to garbage
                if recycled[0]:
                    # gc callback called prematurely
                    t.join(10)
                    return False
                last = RawGreenlet()
                if recycled[0]:
                    break # yes! gc called in green_new
                l.append(last) # increase allocation counter
            else:
                # gc callback not called when expected
                gc.collect()
                if recycled[0]:
                    t.join(10)
                return False
            self.assertEqual(last.parent, current)
            for g in l:
                self.assertEqual(g.parent, current)
            return True
        for _ in range(5):
            if attempt():
                break

    def test_issue_245_reference_counting_subclass_no_threads(self):
        # https://github.com/python-greenlet/greenlet/issues/245
        # Before the fix, this crashed pretty reliably on
        # Python 3.10, at least on macOS; but much less reliably on other
        # interpreters (memory layout must have changed).
        # The threaded test crashed more reliably on more interpreters.
        from greenlet import getcurrent
        from greenlet import GreenletExit

        class Greenlet(RawGreenlet):
            pass

        initial_refs = sys.getrefcount(Greenlet)
        # This has to be an instance variable because
        # Python 2 raises a SyntaxError if we delete a local
        # variable referenced in an inner scope.
        self.glets = [] # pylint:disable=attribute-defined-outside-init

        def greenlet_main():
            try:
                getcurrent().parent.switch()
            except GreenletExit:
                self.glets.append(getcurrent())

        # Before the
        for _ in range(10):
            Greenlet(greenlet_main).switch()

        del self.glets
        self.assertEqual(sys.getrefcount(Greenlet), initial_refs)

    def test_issue_245_reference_counting_subclass_threads(self):
        # https://github.com/python-greenlet/greenlet/issues/245
        from threading import Thread
        from threading import Event

        from greenlet import getcurrent

        class MyGreenlet(RawGreenlet):
            pass

        glets = []
        ref_cleared = Event()

        def greenlet_main():
            getcurrent().parent.switch()

        def thread_main(greenlet_running_event):
            mine = MyGreenlet(greenlet_main)
            glets.append(mine)
            # The greenlets being deleted must be active
            mine.switch()
            # Don't keep any reference to it in this thread
            del mine
            # Let main know we published our greenlet.
            greenlet_running_event.set()
            # Wait for main to let us know the references are
            # gone and the greenlet objects no longer reachable
            ref_cleared.wait(10)
            # The creating thread must call getcurrent() (or a few other
            # greenlet APIs) because that's when the thread-local list of dead
            # greenlets gets cleared.
            getcurrent()

        # We start with 3 references to the subclass:
        # - This module
        # - Its __mro__
        # - The __subclassess__ attribute of greenlet
        # - (If we call gc.get_referents(), we find four entries, including
        #   some other tuple ``(greenlet)`` that I'm not sure about but must be part
        #   of the machinery.)
        #
        # On Python 3.10 it's often enough to just run 3 threads; on Python 2.7,
        # more threads are needed, and the results are still
        # non-deterministic. Presumably the memory layouts are different
        initial_refs = sys.getrefcount(MyGreenlet)
        thread_ready_events = []
        for _ in range(
                initial_refs + 45
        ):
            event = Event()
            thread = Thread(target=thread_main, args=(event,))
            thread_ready_events.append(event)
            thread.start()


        for done_event in thread_ready_events:
            done_event.wait(10)


        del glets[:]
        ref_cleared.set()
        # Let any other thread run; it will crash the interpreter
        # if not fixed (or silently corrupt memory and we possibly crash
        # later).
        self.wait_for_pending_cleanups()
        self.assertEqual(sys.getrefcount(MyGreenlet), initial_refs)

    def test_falling_off_end_switches_to_unstarted_parent_raises_error(self):
        def no_args():
            return 13

        parent_never_started = RawGreenlet(no_args)

        def leaf():
            return 42

        child = RawGreenlet(leaf, parent_never_started)

        # Because the run function takes to arguments
        with self.assertRaises(TypeError):
            child.switch()

    def test_falling_off_end_switches_to_unstarted_parent_works(self):
        def one_arg(x):
            return (x, 24)

        parent_never_started = RawGreenlet(one_arg)

        def leaf():
            return 42

        child = RawGreenlet(leaf, parent_never_started)

        result = child.switch()
        self.assertEqual(result, (42, 24))

    def test_switch_to_dead_greenlet_with_unstarted_perverse_parent(self):
        class Parent(RawGreenlet):
            def __getattribute__(self, name):
                if name == 'run':
                    raise SomeError


        parent_never_started = Parent()
        seen = []
        child = RawGreenlet(lambda: seen.append(42), parent_never_started)
        # Because we automatically start the parent when the child is
        # finished
        with self.assertRaises(SomeError):
            child.switch()

        self.assertEqual(seen, [42])

        with self.assertRaises(SomeError):
            child.switch()
        self.assertEqual(seen, [42])

    def test_switch_to_dead_greenlet_reparent(self):
        seen = []
        parent_never_started = RawGreenlet(lambda: seen.append(24))
        child = RawGreenlet(lambda: seen.append(42))

        child.switch()
        self.assertEqual(seen, [42])

        child.parent = parent_never_started
        # This actually is the same as switching to the parent.
        result = child.switch()
        self.assertIsNone(result)
        self.assertEqual(seen, [42, 24])

    def test_can_access_f_back_of_suspended_greenlet(self):
        # On Python 3.12, they added a ->previous field to
        # _PyInterpreterFrame that has to be cleared when a frame is inactive.
        # If we got that wrong, this immediately crashes.
        main = greenlet.getcurrent()

        def Hub():
            main.switch()

        hub = RawGreenlet(Hub)
        # start it
        hub.switch()
        # now it is suspended
        self.assertIsNotNone(hub.gr_frame)
        # The next line is what would crash
        self.assertIsNone(hub.gr_frame.f_back)




class TestGreenletSetParentErrors(TestCase):
    def test_threaded_reparent(self):
        data = {}
        created_event = threading.Event()
        done_event = threading.Event()

        def run():
            data['g'] = RawGreenlet(lambda: None)
            created_event.set()
            done_event.wait(10)

        def blank():
            greenlet.getcurrent().parent.switch()

        thread = threading.Thread(target=run)
        thread.start()
        created_event.wait(10)
        g = RawGreenlet(blank)
        g.switch()
        with self.assertRaises(ValueError) as exc:
            g.parent = data['g']
        done_event.set()
        thread.join(10)

        self.assertEqual(str(exc.exception), "parent cannot be on a different thread")

    def test_unexpected_reparenting(self):
        another = []
        def worker():
            g = RawGreenlet(lambda: None)
            another.append(g)
            g.switch()
        t = threading.Thread(target=worker)
        t.start()
        t.join(10)
        # The first time we switch (running g_initialstub(), which is
        # when we look up the run attribute) we attempt to change the
        # parent to one from another thread (which also happens to be
        # dead). ``g_initialstub()`` should detect this and raise a
        # greenlet error.
        #
        # EXCEPT: With the fix for #252, this is actually detected
        # sooner, when setting the parent itself. Prior to that fix,
        # the main greenlet from the background thread kept a valid
        # value for ``run_info``, and appeared to be a valid parent
        # until we actually started the greenlet. But now that it's
        # cleared, this test is catching whether ``green_setparent``
        # can detect the dead thread.
        #
        # Further refactoring once again changes this back to a greenlet.error
        #
        # We need to wait for the cleanup to happen, but we're
        # deliberately leaking a main greenlet here.
        self.wait_for_pending_cleanups(initial_main_greenlets=self.main_greenlets_before_test + 1)

        class convoluted(RawGreenlet):
            def __getattribute__(self, name):
                if name == 'run':
                    self.parent = another[0] # pylint:disable=attribute-defined-outside-init
                return RawGreenlet.__getattribute__(self, name)
        g = convoluted(lambda: None)
        with self.assertRaises(greenlet.error) as exc:
            g.switch()
        self.assertEqual(str(exc.exception),
                         "cannot switch to a different thread (which happens to have exited)")
        del another[:]

    def test_unexpected_reparenting_thread_running(self):
        # Like ``test_unexpected_reparenting``, except the background thread is
        # actually still alive.
        another = []
        switched_to_greenlet = threading.Event()
        keep_main_alive = threading.Event()
        def worker():
            g = RawGreenlet(lambda: None)
            another.append(g)
            g.switch()
            switched_to_greenlet.set()
            keep_main_alive.wait(10)
        class convoluted(RawGreenlet):
            def __getattribute__(self, name):
                if name == 'run':
                    self.parent = another[0] # pylint:disable=attribute-defined-outside-init
                return RawGreenlet.__getattribute__(self, name)

        t = threading.Thread(target=worker)
        t.start()

        switched_to_greenlet.wait(10)
        try:
            g = convoluted(lambda: None)

            with self.assertRaises(greenlet.error) as exc:
                g.switch()
            self.assertEqual(str(exc.exception), "cannot switch to a different thread")
        finally:
            keep_main_alive.set()
            t.join(10)
            # XXX: Should handle this automatically.
            del another[:]

    def test_cannot_delete_parent(self):
        worker = RawGreenlet(lambda: None)
        self.assertIs(worker.parent, greenlet.getcurrent())

        with self.assertRaises(AttributeError) as exc:
            del worker.parent
        self.assertEqual(str(exc.exception), "can't delete attribute")

    def test_cannot_delete_parent_of_main(self):
        with self.assertRaises(AttributeError) as exc:
            del greenlet.getcurrent().parent
        self.assertEqual(str(exc.exception), "can't delete attribute")


    def test_main_greenlet_parent_is_none(self):
        # assuming we're in a main greenlet here.
        self.assertIsNone(greenlet.getcurrent().parent)

    def test_set_parent_wrong_types(self):
        def bg():
            # Go back to main.
            greenlet.getcurrent().parent.switch()

        def check(glet):
            for p in None, 1, self, "42":
                with self.assertRaises(TypeError) as exc:
                    glet.parent = p

                self.assertEqual(
                    str(exc.exception),
                    "GreenletChecker: Expected any type of greenlet, not " + type(p).__name__)

        # First, not running
        g = RawGreenlet(bg)
        self.assertFalse(g)
        check(g)

        # Then when running.
        g.switch()
        self.assertTrue(g)
        check(g)

        # Let it finish
        g.switch()


    def test_trivial_cycle(self):
        glet = RawGreenlet(lambda: None)
        with self.assertRaises(ValueError) as exc:
            glet.parent = glet
        self.assertEqual(str(exc.exception), "cyclic parent chain")

    def test_trivial_cycle_main(self):
        # This used to produce a ValueError, but we catch it earlier than that now.
        with self.assertRaises(AttributeError) as exc:
            greenlet.getcurrent().parent = greenlet.getcurrent()
        self.assertEqual(str(exc.exception), "cannot set the parent of a main greenlet")

    def test_deeper_cycle(self):
        g1 = RawGreenlet(lambda: None)
        g2 = RawGreenlet(lambda: None)
        g3 = RawGreenlet(lambda: None)

        g1.parent = g2
        g2.parent = g3
        with self.assertRaises(ValueError) as exc:
            g3.parent = g1
        self.assertEqual(str(exc.exception), "cyclic parent chain")


class TestRepr(TestCase):

    def assertEndsWith(self, got, suffix):
        self.assertTrue(got.endswith(suffix), (got, suffix))

    def test_main_while_running(self):
        r = repr(greenlet.getcurrent())
        self.assertEndsWith(r, " current active started main>")

    def test_main_in_background(self):
        main = greenlet.getcurrent()
        def run():
            return repr(main)

        g = RawGreenlet(run)
        r = g.switch()
        self.assertEndsWith(r, ' suspended active started main>')

    def test_initial(self):
        r = repr(RawGreenlet())
        self.assertEndsWith(r, ' pending>')

    def test_main_from_other_thread(self):
        main = greenlet.getcurrent()

        class T(threading.Thread):
            original_main = thread_main = None
            main_glet = None
            def run(self):
                self.original_main = repr(main)
                self.main_glet = greenlet.getcurrent()
                self.thread_main = repr(self.main_glet)

        t = T()
        t.start()
        t.join(10)

        self.assertEndsWith(t.original_main, ' suspended active started main>')
        self.assertEndsWith(t.thread_main, ' current active started main>')
        # give the machinery time to notice the death of the thread,
        # and clean it up. Note that we don't use
        # ``expect_greenlet_leak`` or wait_for_pending_cleanups,
        # because at this point we know we have an extra greenlet
        # still reachable.
        for _ in range(3):
            time.sleep(0.001)

        # In the past, main greenlets, even from dead threads, never
        # really appear dead. We have fixed that, and we also report
        # that the thread is dead in the repr. (Do this multiple times
        # to make sure that we don't self-modify and forget our state
        # in the C++ code).
        for _ in range(3):
            self.assertTrue(t.main_glet.dead)
            r = repr(t.main_glet)
            self.assertEndsWith(r, ' (thread exited) dead>')

    def test_dead(self):
        g = RawGreenlet(lambda: None)
        g.switch()
        self.assertEndsWith(repr(g), ' dead>')
        self.assertNotIn('suspended', repr(g))
        self.assertNotIn('started', repr(g))
        self.assertNotIn('active', repr(g))

    def test_formatting_produces_native_str(self):
        # https://github.com/python-greenlet/greenlet/issues/218
        # %s formatting on Python 2 was producing unicode, not str.

        g_dead = RawGreenlet(lambda: None)
        g_not_started = RawGreenlet(lambda: None)
        g_cur = greenlet.getcurrent()

        for g in g_dead, g_not_started, g_cur:

            self.assertIsInstance(
                '%s' % (g,),
                str
            )
            self.assertIsInstance(
                '%r' % (g,),
                str,
            )


class TestMainGreenlet(TestCase):
    # Tests some implementation details, and relies on some
    # implementation details.

    def _check_current_is_main(self):
        # implementation detail
        assert 'main' in repr(greenlet.getcurrent())

        t = type(greenlet.getcurrent())
        assert 'main' not in repr(t)
        return t

    def test_main_greenlet_type_can_be_subclassed(self):
        main_type = self._check_current_is_main()
        subclass = type('subclass', (main_type,), {})
        self.assertIsNotNone(subclass)

    def test_main_greenlet_is_greenlet(self):
        self._check_current_is_main()
        self.assertIsInstance(greenlet.getcurrent(), RawGreenlet)



class TestBrokenGreenlets(TestCase):
    # Tests for things that used to, or still do, terminate the interpreter.
    # This often means doing unsavory things.

    def test_failed_to_initialstub(self):
        def func():
            raise AssertionError("Never get here")


        g = greenlet._greenlet.UnswitchableGreenlet(func)
        g.force_switch_error = True

        with self.assertRaisesRegex(SystemError,
                                    "Failed to switch stacks into a greenlet for the first time."):
            g.switch()

    def test_failed_to_switch_into_running(self):
        runs = []
        def func():
            runs.append(1)
            greenlet.getcurrent().parent.switch()
            runs.append(2)
            greenlet.getcurrent().parent.switch()
            runs.append(3) # pragma: no cover

        g = greenlet._greenlet.UnswitchableGreenlet(func)
        g.switch()
        self.assertEqual(runs, [1])
        g.switch()
        self.assertEqual(runs, [1, 2])
        g.force_switch_error = True

        with self.assertRaisesRegex(SystemError,
                                    "Failed to switch stacks into a running greenlet."):
            g.switch()

        # If we stopped here, we would fail the leakcheck, because we've left
        # the ``inner_bootstrap()`` C frame and its descendents hanging around,
        # which have a bunch of Python references. They'll never get cleaned up
        # if we don't let the greenlet finish.
        g.force_switch_error = False
        g.switch()
        self.assertEqual(runs, [1, 2, 3])

    def test_failed_to_slp_switch_into_running(self):
        ex = self.assertScriptRaises('fail_slp_switch.py')

        self.assertIn('fail_slp_switch is running', ex.output)
        self.assertIn(ex.returncode, self.get_expected_returncodes_for_aborted_process())

    def test_reentrant_switch_two_greenlets(self):
        # Before we started capturing the arguments in g_switch_finish, this could crash.
        output = self.run_script('fail_switch_two_greenlets.py')
        self.assertIn('In g1_run', output)
        self.assertIn('TRACE', output)
        self.assertIn('LEAVE TRACE', output)
        self.assertIn('Falling off end of main', output)
        self.assertIn('Falling off end of g1_run', output)
        self.assertIn('Falling off end of g2', output)

    def test_reentrant_switch_three_greenlets(self):
        # On debug builds of greenlet, this used to crash with an assertion error;
        # on non-debug versions, it ran fine (which it should not do!).
        # Now it always crashes correctly with a TypeError
        ex = self.assertScriptRaises('fail_switch_three_greenlets.py', exitcodes=(1,))

        self.assertIn('TypeError', ex.output)
        self.assertIn('positional arguments', ex.output)

    def test_reentrant_switch_three_greenlets2(self):
        # This actually passed on debug and non-debug builds. It
        # should probably have been triggering some debug assertions
        # but it didn't.
        #
        # I think the fixes for the above test also kicked in here.
        output = self.run_script('fail_switch_three_greenlets2.py')
        self.assertIn(
            "RESULTS: [('trace', 'switch'), "
            "('trace', 'switch'), ('g2 arg', 'g2 from tracefunc'), "
            "('trace', 'switch'), ('main g1', 'from g2_run'), ('trace', 'switch'), "
            "('g1 arg', 'g1 from main'), ('trace', 'switch'), ('main g2', 'from g1_run'), "
            "('trace', 'switch'), ('g1 from parent', 'g1 from main 2'), ('trace', 'switch'), "
            "('main g1.2', 'g1 done'), ('trace', 'switch'), ('g2 from parent', ()), "
            "('trace', 'switch'), ('main g2.2', 'g2 done')]",
            output
        )

    def test_reentrant_switch_GreenletAlreadyStartedInPython(self):
        output = self.run_script('fail_initialstub_already_started.py')

        self.assertIn(
            "RESULTS: ['Begin C', 'Switch to b from B.__getattribute__ in C', "
            "('Begin B', ()), '_B_run switching to main', ('main from c', 'From B'), "
            "'B.__getattribute__ back from main in C', ('Begin A', (None,)), "
            "('A dead?', True, 'B dead?', True, 'C dead?', False), "
            "'C done', ('main from c.2', None)]",
            output
        )

    def test_reentrant_switch_run_callable_has_del(self):
        output = self.run_script('fail_clearing_run_switches.py')
        self.assertIn(
             "RESULTS ["
            "('G.__getattribute__', 'run'), ('RunCallable', '__del__'), "
            "('main: g.switch()', 'from RunCallable'), ('run_func', 'enter')"
            "]",
            output
        )

if __name__ == '__main__':
    import unittest
    unittest.main()



from ._SitefFunction import (
    SitefFunctions,
    PaymentCodes,
    CancelCodes,
    ResultCodes,
    CommandCodes,
    FieldCodes,
    FinalizeActionFunction,
)

__all__ = [
    "SitefFunctions",
    "CancelCodes",
    "PaymentCodes",
    "ResultCodes",
    "CommandCodes",
    "FieldCodes",
    "FinalizeActionFunction",
]



import logging
from threading import Lock

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from application.configurator import Configurator
from application.interactor import (
    CatalogUpdaterInteractor,
    MediaUpdaterInteractor,
    UserUpdaterInteractor,
    GetCatalogVersionToApplyInteractor,
    LoaderUpdaterInteractor,
    RemoteLoaderUpdaterInteractor,
    BinaryUpdaterInteractor,
    GetVersionsInteractor,
    ChromeUpdaterInteractor,
    HasPendingLoaderUpdateInteractor,
)
from application.manager import (
    CatalogUpdaterManager,
    UserUpdaterManager,
    LoaderUpdaterManager,
    BinaryUpdaterManager,
    MediaUpdaterManager,
)
from application.model import (
    ListenedTokens,
    ListenedEvents,
    UpdateType,
    AsyncUuidMessageProcessor,
    Updater,
)
from application.model.customexception import (
    DisabledFunctionality,
    NoUpdateFound,
    HasNoProductsImagesToDownload,
)
from application.processor import (
    CatalogUpdaterProcessor,
    MediaUpdaterProcessor,
    GetCatalogVersionToApplyProcessor,
    UserUpdaterProcessor,
    LoaderUpdaterProcessor,
    RemoteLoaderUpdaterProcessor,
    BinaryUpdaterProcessor,
    GetVersionsProcessor,
    GetBinaryUpdaterProcessor,
    ChromeUpdaterProcessor,
    HasPendingLoaderUpdateProcessor,
)
from application.repository.api import (
    CatalogUpdaterAPIRepository,
    MediaUpdaterAPIRepository,
    LoaderUpdaterAPIRepository,
    UserUpdaterAPIRepository,
    BinaryUpdaterAPIRepository,
)
from application.repository.pos import (
    CatalogUpdaterPOSRepository,
    CommonPOSRepository,
    BinaryUpdaterPOSRepository,
)
from application.service import (
    StartStopUpdaterService,
    ChromeUpdaterService,
    UpdaterService,
    LoaderUpdaterService,
    ApacheUpdaterService,
    MediaUpdaterService,
)
from application.util import LoggerUtil
from messagehandler import MessageHandlerBuilder
from messageprocessor import (
    MessageProcessorMessageHandler,
    DefaultMessageProcessorExecutorFactory,
)
from messageprocessorutil import LoggingProcessorCallback
from mwhelper import BaseRepository

from messagebus import Message, Event

logger = LoggerUtil.get_logger_name(__name__)


class ComponentMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, configurator, pos_repositories, common_pos_repository):
        # type: (Configurator, List[BaseRepository], CommonPOSRepository) -> None

        self.logger = logger
        self.configs = configurator.configs
        self.pos_repositories = pos_repositories
        self.common_pos_repository = common_pos_repository

        self.message_handler = None  # type: Union[None, MessageProcessorMessageHandler]
        self.updaters = {}  # type: Dict[str: Updater]
        self.working_on_update = Lock()
        self.binary_updater_interactor = None  # type: Optional[BinaryUpdaterInteractor]

    def build_singletons(self):
        # type: () -> None

        version_to_apply = ListenedTokens.TK_POS_GET_CATALOG_VERSION_TO_APPLY.value
        get_versions = ListenedTokens.TK_POS_UPDATER_GET_VERSIONS.value

        api_repository = CatalogUpdaterAPIRepository(self.logger, self.configs)
        pos_repository = self.pos_repositories[UpdateType.catalog]
        updater_service = UpdaterService(
            api_repository,
            pos_repository,
            self.common_pos_repository,
            self.working_on_update,
            self.configs
        )

        catalog_enabled = self.configs.updaters[UpdateType.catalog.name].enabled
        catalog_interactor = GetCatalogVersionToApplyInteractor(pos_repository, updater_service, catalog_enabled)
        get_catalog_version_to_apply_interactor = catalog_interactor

        get_version_repository = self.pos_repositories["GetVersions"]
        get_versions_interactor = GetVersionsInteractor(
            pos_repository=get_version_repository,
            path_to_catalog_json=self.configs.path_to_catalog_json,
            logger=self.logger,
        )

        has_pending_loader_update_interactor = HasPendingLoaderUpdateInteractor(
            pos_repository=self.pos_repositories[UpdateType.loader],
        )

        has_pending_loader_update_processor = HasPendingLoaderUpdateProcessor(
            interactor=has_pending_loader_update_interactor,
        )

        msg_processors = {
            version_to_apply: GetCatalogVersionToApplyProcessor(get_catalog_version_to_apply_interactor),
            get_versions: GetVersionsProcessor(get_versions_interactor)
        }

        binaries_repository = self.pos_repositories[UpdateType.binaries]

        media_updater_processor, media_updater_service = self._build_media_singletons(
            binaries_repository=binaries_repository,
        )
        catalog_updater_processor = self._build_catalog_singletons(
            pos_repo=pos_repository,
            updater_service=updater_service,
            api_repo=api_repository,
            media_updater_service=media_updater_service,
            binaries_repository=binaries_repository,
        )
        user_updater_processor = self._build_user_singletons(
            binaries_repository=binaries_repository,
        )
        apache_updater_service = self._build_apache_updater_singletons()
        loader_updater_processor, loader_remote_processor, loader_updater_service = self._build_loader_singletons(
            binaries_repository=binaries_repository,
        )
        chrome_updater_processor, chrome_updater_service = self._build_chrome_updater_singletons()
        start_stop_updater_service = self._build_start_stop_updater_singletons()
        binary_updater_processor, get_binary_update_processor = self._build_binary_singletons(
            binaries_repository=binaries_repository,
            loader_updater_service=loader_updater_service,
            chrome_updater_service=chrome_updater_service,
            start_stop_updater_service=start_stop_updater_service,
            apache_updater_service=apache_updater_service,
        )

        event_processors = {}
        if catalog_updater_processor:
            msg_processors[ListenedTokens.TK_POS_UPDATER_PERFORM_CATALOG_UPDATE.value] = catalog_updater_processor
        if media_updater_processor:
            msg_processors[ListenedTokens.TK_POS_UPDATER_PERFORM_MEDIA_UPDATE.value] = media_updater_processor
            event_processors[ListenedEvents.EVT_POS_UPDATER_PERFORM_MEDIA_UPDATE.value] = media_updater_processor
        if user_updater_processor:
            msg_processors[ListenedTokens.TK_POS_UPDATER_PERFORM_USER_UPDATE.value] = user_updater_processor
            event_processors[ListenedEvents.EVT_POS_UPDATER_PERFORM_USER_UPDATE.value] = user_updater_processor
        if loader_updater_processor:
            msg_processors[ListenedTokens.TK_POS_UPDATER_PERFORM_LOADER_UPDATE.value] = loader_updater_processor
            event_processors[ListenedEvents.EVT_POS_UPDATER_PERFORM_LOADER_UPDATE.value] = loader_updater_processor
        if loader_remote_processor:
            event_processors[ListenedEvents.EVT_POS_UPDATER_REMOTE_LOADER_UPDATE.value] = loader_remote_processor
        if binary_updater_processor:
            msg_processors[ListenedTokens.TK_POS_UPDATER_PERFORM_BINARY_UPDATE.value] = binary_updater_processor
            event_processors[ListenedEvents.EVT_POS_UPDATER_PERFORM_BINARY_UPDATE.value] = binary_updater_processor
        if get_binary_update_processor:
            msg_processors[ListenedTokens.TK_POS_HAS_BINARY_UPDATE_TO_APPLY.value] = get_binary_update_processor
        if chrome_updater_processor:
            msg_processors[ListenedTokens.TK_POS_UPDATER_PERFORM_CHROME_UPDATE.value] = chrome_updater_processor

        pending_loader_update_token = ListenedTokens.TK_POS_UPDATER_HAS_PENDING_LOADER_UPDATE.value
        msg_processors[pending_loader_update_token] = has_pending_loader_update_processor

        log_level_exceptions = {
            DisabledFunctionality: logging.WARN,
            NoUpdateFound: logging.INFO,
            HasNoProductsImagesToDownload: logging.INFO
        }
        callbacks = [LoggingProcessorCallback(self.logger, log_level_exceptions)]

        self.message_handler = MessageProcessorMessageHandler(
            event_processors,
            msg_processors,
            DefaultMessageProcessorExecutorFactory(callbacks)
        )

        self._start_managers()

    def _start_managers(self):
        for updater in self.updaters:
            if self.updaters[updater].manager:
                self.updaters[updater].manager.start()

    def build_message_handler(self, message, event):
        # type: (Message, Event) -> None

        return self.message_handler

    def destroy_singletons(self):
        # type: () -> None

        for updater in self.updaters:
            if self.updaters[updater].manager:
                self.updaters[updater].manager.stop()

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandlerBuilder) -> None

        return

    def _build_catalog_singletons(
        self,
        pos_repo,               # type: CatalogUpdaterPOSRepository
        updater_service,        # type: UpdaterService
        api_repo,               # type: CatalogUpdaterAPIRepository
        media_updater_service,  # type: MediaUpdaterService
        binaries_repository,    # type: BinaryUpdaterPOSRepository
    ):
        # type: (...) -> Optional[CatalogUpdaterProcessor] # noqa

        update_type = UpdateType.catalog
        update_name = update_type.name

        if update_name not in self.configs.updaters:
            return

        local_configs = self.configs.updaters[update_name]
        interactor = CatalogUpdaterInteractor(self.configs, self.working_on_update, api_repo, pos_repo, updater_service,
                                              local_configs.enabled)
        processor = CatalogUpdaterProcessor(interactor)
        manager = CatalogUpdaterManager(
            configs=self.configs,
            interactor=interactor,
            api_repository=api_repo,
            pos_repository=pos_repo,
            updater_service=updater_service,
            media_updater_service=media_updater_service,
            binaries_repository=binaries_repository,
            common_pos_repository=self.common_pos_repository,
        )

        local_configs = self.configs.updaters[update_type.name]

        self.updaters[local_configs.name] = Updater(update_type=update_type,
                                                    manager=manager,
                                                    interactor=interactor)

        return processor

    def _build_media_singletons(
        self,
        binaries_repository,    # type: BinaryUpdaterPOSRepository
    ):
        # type: (...) -> Optional[Tuple[Union[AsyncUuidMessageProcessor, None], MediaUpdaterService]]

        update_type = UpdateType.media
        update_name = update_type.name

        if update_name not in self.configs.updaters:
            return

        update_type = UpdateType.media
        update_name = update_type.name

        if update_name not in self.configs.updaters:
            return

        local_configs = self.configs.updaters[update_name]
        api_repo = MediaUpdaterAPIRepository(self.configs)
        pos_repo = self.pos_repositories[UpdateType.media]
        enabled = local_configs.enabled

        updater_service = UpdaterService(
            api_repo,
            pos_repo,
            self.common_pos_repository,
            self.working_on_update,
            self.configs
        )

        media_updater_service = MediaUpdaterService(
            configs=self.configs,
            working_on_update=self.working_on_update,
            api_repository=api_repo,
            pos_repository=pos_repo,
            updater_service=updater_service,
            enabled=enabled,
        )

        local_configs = self.configs.updaters[update_name]
        interactor = MediaUpdaterInteractor(media_updater_service=media_updater_service)
        processor = MediaUpdaterProcessor(interactor)
        manager = MediaUpdaterManager(
            self.configs,
            interactor,
            pos_repo,
            binaries_repository,
        )

        self.updaters[local_configs.name] = Updater(update_type, manager, interactor)

        return processor, media_updater_service

    def _build_user_singletons(
        self,
        binaries_repository,    # type: BinaryUpdaterPOSRepository
    ):
        # type: (...) -> Union[AsyncUuidMessageProcessor, None]

        update_type = UpdateType.user
        update_name = update_type.name

        if update_name not in self.configs.updaters:
            return

        local_configs = self.configs.updaters[update_name]
        api_repo = UserUpdaterAPIRepository(self.configs)
        pos_repo = self.pos_repositories[UpdateType.user]
        enabled = local_configs.enabled
        interactor = UserUpdaterInteractor(self.configs, self.working_on_update, api_repo, pos_repo, enabled)
        processor = UserUpdaterProcessor(interactor)
        manager = UserUpdaterManager(
            self.configs,
            interactor,
            pos_repo,
            binaries_repository,
        )

        self.updaters[local_configs.name] = Updater(
            update_type=update_type,
            manager=manager,
            interactor=interactor
        )

        return processor

    def _build_loader_singletons(
        self,                 # type: ComponentMessageHandlerBuilder
        binaries_repository,  # type: BinaryUpdaterPOSRepository
    ):
        # type: (...) -> (Union[AsyncUuidMessageProcessor, None], LoaderUpdaterService)

        update_type = UpdateType.loader
        update_name = update_type.name

        if update_name not in self.configs.updaters:
            return

        pos_repo = self.pos_repositories[UpdateType.loader]
        api_repo = LoaderUpdaterAPIRepository(
            logger=logger,
            configs=self.configs,
        )
        updater_service = UpdaterService(
            api_repo,
            pos_repo,
            self.common_pos_repository,
            self.working_on_update,
            self.configs
        )

        catalog_repo = self.pos_repositories[UpdateType.catalog]
        version_repo = self.pos_repositories["GetVersions"]

        loader_updater_service = LoaderUpdaterService(
            self.configs,
            self.working_on_update,
            api_repo,
            pos_repo,
            version_repo,
            updater_service
        )
        remote_updater_interactor = RemoteLoaderUpdaterInteractor(
            loader_updater_service=loader_updater_service,
            remote_update_window=self.configs.updaters.get("loader").remote_update_window,
            working_on_update=self.working_on_update,
            pos_repository=pos_repo,
        )
        remote_updater_processor = RemoteLoaderUpdaterProcessor(interactor=remote_updater_interactor)
        loader_updater_interactor = LoaderUpdaterInteractor(loader_updater_service=loader_updater_service)
        loader_updater_processor = LoaderUpdaterProcessor(interactor=loader_updater_interactor)
        manager = LoaderUpdaterManager(
            configs=self.configs,
            interactor=loader_updater_interactor,
            pos_repository=pos_repo,
            catalog_pos_repository=catalog_repo,
            binaries_repository=binaries_repository,
            loader_updater_service=loader_updater_service,
        )

        local_configs = self.configs.updaters[update_name]
        self.updaters[local_configs.name] = Updater(
            update_type=update_type,
            manager=manager,
            interactor=loader_updater_interactor,
        )

        return loader_updater_processor, remote_updater_processor, loader_updater_service

    def _build_apache_updater_singletons(self):
        # type: () -> ApacheUpdaterService

        pos_repo = self.pos_repositories[UpdateType.binaries]
        updater_service = UpdaterService(
            api_repository=None,
            pos_repository=pos_repo,
            common_repository=self.common_pos_repository,
            working_on_update=self.working_on_update,
            configs=self.configs,
        )

        return ApacheUpdaterService(updater_service=updater_service)

    def _build_chrome_updater_singletons(
        self
    ):
        # type: () -> ChromeUpdaterProcessor

        pos_repo = self.pos_repositories[UpdateType.binaries]
        updater_service = UpdaterService(
            api_repository=None,
            pos_repository=pos_repo,
            common_repository=self.common_pos_repository,
            working_on_update=self.working_on_update,
            configs=self.configs
        )

        chrome_updater_service = ChromeUpdaterService(
            configs=self.configs,
            working_on_update=self.working_on_update,
            updater_service=updater_service,
            pos_repository=pos_repo
        )

        chrome_updater_interactor = ChromeUpdaterInteractor(
            chrome_updater_service=chrome_updater_service
        )

        return ChromeUpdaterProcessor(chrome_updater_interactor), chrome_updater_service

    def _build_start_stop_updater_singletons(
        self
    ):
        # type: () -> StartStopUpdaterService

        pos_repo = self.pos_repositories[UpdateType.binaries]
        updater_service = UpdaterService(
            api_repository=None,
            pos_repository=pos_repo,
            common_repository=self.common_pos_repository,
            working_on_update=self.working_on_update,
            configs=self.configs,
        )

        start_stop_updater_service = StartStopUpdaterService(updater_service=updater_service)

        return start_stop_updater_service

    def _build_binary_singletons(
        self,
        binaries_repository,            # type: BinaryUpdaterPOSRepository
        loader_updater_service,         # type: LoaderUpdaterService
        chrome_updater_service,         # type: ChromeUpdaterService
        start_stop_updater_service,     # type: StartStopUpdaterService
        apache_updater_service,         # type: ApacheUpdaterService
    ):
        # type: (...) -> Union[AsyncUuidMessageProcessor, None]

        update_type = UpdateType.binaries
        update_name = update_type.name

        if update_name not in self.configs.updaters:
            return

        api_repo = BinaryUpdaterAPIRepository(self.configs)
        updater_service = UpdaterService(
            api_repo,
            binaries_repository,
            self.common_pos_repository,
            self.working_on_update,
            self.configs
        )
        interactor = BinaryUpdaterInteractor(
            configs=self.configs,
            working_on_update=self.working_on_update,
            api_repository=api_repo,
            pos_repository=binaries_repository,
            updater_service=updater_service,
            loader_updater_service=loader_updater_service,
            chrome_updater_service=chrome_updater_service,
            start_stop_updater_service=start_stop_updater_service,
            apache_updater_service=apache_updater_service,
        )
        self.binary_updater_interactor = interactor
        self.binary_updater_interactor.update_binaries_start()
        processor = BinaryUpdaterProcessor(interactor)
        manager = BinaryUpdaterManager(
            self.configs,
            interactor,
            binaries_repository
        )
        has_update_processor = GetBinaryUpdaterProcessor(interactor)
        local_configs = self.configs.updaters[update_name]
        self.updaters[local_configs.name] = Updater(update_type, manager, interactor)

        return processor, has_update_processor



from production.box import CustomParamCourseBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class CustomParamCourseBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, publish_scheduler, loggers):
        super(CustomParamCourseBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository
        self.publish_scheduler = publish_scheduler

    def build(self, config):
        # type: (BoxConfiguration) -> CustomParamCourseBox

        return CustomParamCourseBox(
                config.name,
                config.sons,
                self.product_repository,
                self.publish_scheduler,
                self.get_logger(config)
        )

    @staticmethod
    def course_comparer(c1, c2):
        return int(c1[2]) - int(c2[2])



# -*- coding: utf-8 -*-
import inspect

from xml.etree import cElementTree as eTree
from functools import wraps
from datetime import datetime

from posot import OrderTaker
from sysactions import (
    StopAction,
    UiOutOfSync,
    get_current_order,
    get_line_product_name,
    get_model,
    get_operator_session,
    get_podtype,
    get_posot,
    show_messagebox,
)
from systools import sys_log_exception

from actions import logger
from actions.config import get_function_authorization_level_from_config
from actions.util import (
    get_user_level,
    get_authorization,
)
from actions.order import (
    must_sync_ui,
    sync_ui_and_model,
)
from actions.pos import pos_need_to_be_blocked
from api.order import get_order_service

from typing import (
    Any,
    Dict,
    List,
    Optional,
)

configurations = None


def initialize_manager_authorizations(config):
    global configurations
    configurations = config


def user_authorized_to_execute_action(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        user_id = None
        pos_id = _get_pos_id(args, kwargs)
        if not pos_id:
            return False

        try:
            function_level = get_function_authorization_level_from_config(
                config=configurations,
                function_name=f.func_name,
            )
            model = get_model(posid=pos_id)
            operator = model.find("./Operator")
            if operator is not None:
                user_id = operator.get("id")
                if get_user_level(user_id=user_id) <= function_level:
                    status, user_id = get_authorization(
                        pos_id=pos_id,
                        min_level=function_level,
                        model=model,
                    )
                    if not status:
                        return StopAction()

                if function_level > 0:
                    _save_authentication_info_on_order(
                        executed_function=f.func_name,
                        model=model,
                        pos_id=pos_id,
                        user_id=user_id,
                    )

        except (Exception,) as ex:
            error_message = "$GET_USER_AUTHORIZATION_ERROR_" + str(type(ex).__name__).upper()
            sys_log_exception("Error validating user authorization")
            show_messagebox(pos_id, message=error_message, icon="error")
            return False

        if 'auth_data' in inspect.getargspec(f).args:
            kwargs['auth_data'] = user_id

        return f(*args, **kwargs)

    return decorated_function


def block_action_if_pos_is_blocked(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        pos_id = _get_pos_id(args, kwargs)
        if not pos_id:
            return False

        try:
            model = get_model(pos_id)
            pod_type = get_podtype(model)
            if pod_type != 'TT' and pos_need_to_be_blocked(pos_id):
                show_messagebox(pos_id, message="$POS_BLOCKED", icon="error")
                raise StopAction()
        except (StopAction,):
            raise
        except (Exception,):
            return False

        return f(*args, **kwargs)

    return decorated_function


def block_action_if_there_is_no_order(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        pos_id = _get_pos_id(args, kwargs)
        if not pos_id:
            return False

        if not get_order_service(pos_id=pos_id).has_current_order():
            show_messagebox(pos_id, "$HAS_NO_ITEMS_IN_ORDER", title="")
            return False

        return f(*args, **kwargs)

    return decorated_function


def block_action_if_order_items_not_resolved(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        from posactions import list_open_options

        pos_id = _get_pos_id(args, kwargs)
        if not pos_id:
            return False

        order = get_order_service(pos_id=pos_id).get_order_picture()

        sale_lines = order.findall("SaleLine")
        active_sale_lines = _get_active_order_lines(sale_lines)
        is_remote_order = order.find("CustomOrderProperties/OrderProperty[@key='REMOTE_ORDER_ID']")
        if not active_sale_lines and is_remote_order is None:
            show_messagebox(pos_id, "$HAS_NO_ITEMS_IN_ORDER", title="")
            return False

        option = list_open_options(order)
        if option is not None:
            model = get_model(pos_id)
            prod_name = get_line_product_name(model, int(option.get("lineNumber")))
            show_messagebox(pos_id, message="$NEED_TO_RESOLVE_OPTION|%s" % (prod_name.encode("UTF-8")), icon="error")
            return False

        return f(*args, **kwargs)

    return decorated_function


def validate_order_state(
    f,  # type: Any
):
    # type: (...) -> Any[Any, bool]

    @wraps(f)
    def decorated_function(
        *args,      # type: str
        **kwargs    # type: str
    ):
        # type: (...) -> Any[Any, bool]

        pos_id = _get_pos_id(
            args=args,
            kwargs=kwargs,
        )
        if not pos_id:
            return False

        try:
            model = get_model(posid=pos_id)
            order_taker = get_posot(model=model)
            order_id = model.find(".//Order").get("orderId")
            fix_current_order_in_model(
                pos_id=pos_id,
                model=model,
                order_taker=order_taker,
                order_id=order_id,
            )
        except (StopAction, UiOutOfSync):
            raise
        except (Exception,):
            return False

        return f(
            *args,
            **kwargs
        )

    return decorated_function


def fix_current_order_in_model(
    pos_id,         # type: str
    model,          # type: eTree.Element
    order_taker,    # type: OrderTaker
    order_id,       # type: str
):
    # type: (...) -> Optional[bool]

    logger.info("Trying to fix POS {}. Current Order Id: {}".format(
        pos_id,
        order_id,
    ))
    model_order_picture = get_current_order(model=model)
    current_order_picture = eTree.XML(order_taker.orderPicture(orderid=order_id))

    if model_order_picture.get("state") != current_order_picture.find("Order").get("state"):
        sync_ui_and_model(pos_id=pos_id)
    else:
        must_sync_ui(
            pos_id=pos_id,
            different_than="IN_PROGRESS",
        )

    return False


def user_must_be_logged(
    f,  # type: Any
):
    # type: (...) -> Any[Any, bool]

    @wraps(f)
    def decorated_function(
        *args,      # type: List[Any]
        **kwargs    # type: Dict[str, Any]
    ):
        # type: (...) -> Any[Any, bool]

        pos_id = _get_pos_id(
            args=args,
            kwargs=kwargs,
        )
        if not pos_id:
            return False

        try:
            model = get_model(posid=pos_id)
            session_id = get_operator_session(model=model)
            if session_id in ("", None, "None"):
                message = "$USER_MUST_BE_LOGGED_TO_DO_THE_REQUIRED_ACTION_AND_UI_WILL_BE_REFRESH"
                raise UiOutOfSync(
                    pos_id=pos_id,
                    custom_message=message,
                )
        except (StopAction, UiOutOfSync):
            raise
        except (Exception,):
            return False

        return f(
            *args,
            **kwargs
        )

    return decorated_function


def _get_pos_id(
    args,       # type: List[Any]
    kwargs,     # type: Dict[str, Any]
):
    pos_id = None
    if len(args) > 0:
        pos_id = args[0]
    elif "pos_id" in kwargs:
        pos_id = kwargs["pos_id"]
    elif "posid" in kwargs:
        pos_id = kwargs["posid"]
    else:
        sys_log_exception(message="PosId not found to authorize")

    return pos_id


def _get_order_deleted_lines(
    sale_lines,     # type: List[eTree.Element]
):
    # type: (...) -> List[str]

    return map(
        lambda x: x.get("lineNumber"),
        filter(lambda x: x.get("level") == "0" and x.get("qty") == "0", sale_lines),
    )


def _get_active_order_lines(
    sale_lines,     # type: List[eTree.Element]
):
    # type: (...) -> List[eTree.Element]

    deleted_line_numbers = _get_order_deleted_lines(sale_lines=sale_lines)
    return filter(lambda x: x.get("lineNumber") not in deleted_line_numbers, sale_lines)


def _save_authentication_info_on_order(
    executed_function,      # type: str
    model,                  # type: eTree.Element
    pos_id,                 # type: int
    user_id,                # type: str
):
    # type: (...) -> None

    if executed_function is None:
        return

    if not model:
        model = get_model(posid=pos_id)

    pos_ot = get_posot(model=model)
    order = model.find("CurrentOrder/Order")
    order_id = order.get("orderId") if order else None

    if order_id:
        property_name = "{}: {}".format(executed_function, datetime.now().strftime("%m/%d/%Y %H:%M:%S"))
        pos_ot.setOrderCustomProperty(
            key=property_name,
            value=user_id,
            orderid=order_id,
        )



# bsn.py - functions for handling BSNs
#
# Copyright (C) 2010-2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""BSN (Burgerservicenummer, the Dutch citizen identification number).

The BSN is a unique personal identifier and has been introduced as the
successor to the sofinummer. It is issued to each Dutch national. The number
consists of up to nine digits (the leading zeros are commonly omitted) and
contains a simple checksum.

More information:

* https://en.wikipedia.org/wiki/National_identification_number#Netherlands
* https://nl.wikipedia.org/wiki/Burgerservicenummer
* https://www.burgerservicenummer.nl/

>>> validate('1112.22.333')
'111222333'
>>> validate('1112.52.333')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1112223334')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('111222333')
'1112.22.333'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -.').strip().zfill(9)


def checksum(number):
    """Calculate the checksum over the number. A valid number should have
    a checksum of 0."""
    return (sum((9 - i) * int(n) for i, n in enumerate(number[:-1])) -
            int(number[-1])) % 11


def validate(number):
    """Check if the number is a valid BSN. This checks the length and whether
    the check digit is correct."""
    number = compact(number)
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if checksum(number) != 0:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid BSN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the passed number to the standard presentation format."""
    number = compact(number)
    return number[:4] + '.' + number[4:6] + '.' + number[6:]



import requests
from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import ErrorRetrievingUsers
from helper import retry, json_serialize
from requests import Response
from typing import Dict
from application.util import LoggerUtil, ErrorResponseUtil

logger = LoggerUtil.get_logger_name(__name__)


class UserUpdaterAPIRepository(object):
    def __init__(self, configs):
        # type: (Configurations) -> None

        self.logger = logger
        self.configs = configs
        self.update_type = UpdateType.user
        self.local_configs = configs.updaters[self.update_type.name]
        self.response_util = ErrorResponseUtil()
        self.error_response = None

    def get_employees_data(self):
        # type: () -> Dict

        self.logger.info("Obtaining users data...")

        url = self._get_employee_url()
        headers = self._get_app_json_headers()

        response = retry(
                logger=self.logger,
                method_name="get_employees_data",
                n=3,
                sleep_seconds=5,
                trying_to_do=lambda: requests.request("GET", url=url, headers=headers, timeout=60),
                is_success=lambda resp: resp.status_code == 200,
                success=lambda resp: resp,
                failed=lambda resp: self._get_error_response(resp)
        )

        if response is None:
            raise ErrorRetrievingUsers(self.error_response)

        self.logger.info("Employees data successfully obtained!")

        return response.json()

    def _get_app_json_headers(self):
        # type: () -> Dict[str, str]

        return {"X-Api-Key": self.configs.api_key}

    def _get_employee_url(self):
        # type: () -> str

        end_point_url = self.local_configs.endpoints.employee.format(storeId=self.configs.store_id)
        return self.configs.base_url + end_point_url

    def _get_error_response(self, resp):
        # type: (Response) -> None

        self.logger.error("Error downloading users: {}".format(json_serialize(resp)))
        self.error_response = self.response_util.get_error_response(resp, "$USER_NOT_FOUND")



from production.box import FromOrderStateFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class FromOrderStateFilterBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(FromOrderStateFilterBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> FromOrderStateFilterBox
        return FromOrderStateFilterBox(config.name,
                                       config.sons,
                                       self.get_extra(config, "OrderStates", None),
                                       self.get_logger(config))



from typing import (
    Any,
    List,
)

from devicemanager.interactor.dto import DeviceStatusDto
from devicemanager.interactor.get_devices import GetDevicesInteractor
from devicemanager.processor._DeviceStatusDtoProcessor import DeviceStatusDtoProcessor


class GetDevicesProcessor(DeviceStatusDtoProcessor):
    def __init__(
        self,
        get_devices_interactor,     # type: GetDevicesInteractor
    ):
        # type: (...) -> None

        super(GetDevicesProcessor, self).__init__()
        self.get_devices_interactor = get_devices_interactor

    def parse_data(
        self,           # type: GetDevicesProcessor
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> None

        pass

    def call_business(
        self,   # type: GetDevicesProcessor
        _,      # type: Any
    ):
        # type: (...) -> List[DeviceStatusDto]

        return self.get_devices_interactor.get_devices()



from collections import defaultdict


class CustomStorage(object):
    __slots__ = ['_storage', 'full_storage']

    def __init__(self):
        self._storage = defaultdict(dict)
        self.full_storage = defaultdict(dict)

    def clear_data(self):
        self.__init__()

    def write_data(self, context, key, value, write_to_xml=True):
        self.full_storage[context][key] = value
        if write_to_xml:
            self._storage[context][key] = value

    def read_data(self, context, key):
        dic = self.full_storage[context]
        if hasattr(key, "__iter__"):
            # "key" is actually a list of keys to retrieve at once
            return map(dic.get, key)
        # single key to retrieve
        return dic.get(key)

    def delete_data(self, context, key):
        try:
            del self._storage[context][key]
            del self.full_storage[context][key]
        except KeyError:
            pass

    def locate_data(self, key):
        result = []
        for context in self._storage:
            if key in self._storage[context]:
                result.append(context)

        return result



import json
from typing import Optional
from uuid import UUID

from event_manager_api.shared import EventPublishDetailDto
from event_manager_api.shared.error import (
    NOT_A_UUID,
    INVALID_UTF8_DATA,
)
from event_manager_api.shared.token import TK_EVENT_MANAGER_DEFINITION_NOT_FOUND
from messagebus import (
    AckMessage,
    Message,
    DataType,
    BadRequestMessage,
    InternalErrorMessage,
)
from messageprocessorutil import UuidMessageProcessor
from model.mapper import EventPublishDetailMapper
from model.repository import EventRepository
from .._ParseException import ParseException


class GetDetailsProcessor(UuidMessageProcessor):
    def __init__(self, event_repository: EventRepository):
        super().__init__()
        self.event_repository = event_repository

    def parse_data(self, data):
        try:
            return UUID(data.decode("utf-8"))
        except UnicodeDecodeError:
            raise ParseException(
                error_code=INVALID_UTF8_DATA[0],
                error_message=INVALID_UTF8_DATA[1],
            )
        except ValueError:
            raise ParseException(
                error_code=NOT_A_UUID[0],
                error_message=NOT_A_UUID[1],
            )

    def call_business(self, definition_id: UUID):
        definition = self.event_repository.get_by_id(id=definition_id)
        if definition is None:
            return None
        return EventPublishDetailMapper.map(publish_detail=definition)

    def format_response(
        self,
        message_bus,
        message,
        event,
        input_obj,
        dto: Optional[EventPublishDetailDto],
    ):
        if dto is None:
            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=TK_EVENT_MANAGER_DEFINITION_NOT_FOUND,
                    data_type=DataType.empty,
                )
            )
        else:
            message_bus.reply_message(
                message=message,
                reply_message=AckMessage(data=json.dumps(dto.to_dict()).encode("utf-8"))
            )

    def format_exception(
        self,
        message_bus,
        message,
        event,
        input_obj,
        exception
    ):
        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(ex=exception)
        )

    def format_parse_exception(
        self,
        message_bus,
        message,
        event,
        data,
        exception
    ):
        if isinstance(exception, ParseException):
            message_bus.reply_message(
                message=message,
                reply_message=BadRequestMessage(
                    error_code=exception.error_code,
                    error_message=exception.error_message,
                )
            )
        else:
            self.format_exception(
                message_bus=message_bus,
                message=message,
                event=event,
                input_obj=data,
                exception=exception,
            )



from logging import Logger

from production.repository import ProductRepository
from timeutil import Clock
from typing import List, Optional, Union, Any

from ._PrepTimeBox import PrepTimeBox
from ._ProductionBox import ProductionBox


class CustomParamPrepTimeBox(PrepTimeBox):
    def __init__(self, name, sons, product_repository, clock, publish_scheduler=None, wait_even_if_done=False, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], ProductRepository, Clock, Any, bool, Logger) -> None # noqa
        prep_times = product_repository.get_prep_times()

        super(CustomParamPrepTimeBox, self).__init__(
            name,
            sons,
            prep_times,
            clock,
            publish_scheduler,
            wait_even_if_done,
            logger
        )



cimport cython
from cpython.long cimport PyLong_FromLongLong
from cpython.set cimport PySet_Add

from collections.abc import Collection
from itertools import filterfalse

cdef bint add_not_present(set seen, object item, hashfunc):
    hash_value = hashfunc(item)
    if hash_value not in seen:
        PySet_Add(seen, hash_value)
        return True
    else:
        return False

cdef list cunique_list(seq, hashfunc=None):
    cdef set seen = set()
    if not hashfunc:
        return [x for x in seq if x not in seen and not PySet_Add(seen, x)]
    else:
        return [x for x in seq if add_not_present(seen, x, hashfunc)]

def unique_list(seq, hashfunc=None):
    return cunique_list(seq, hashfunc)

cdef class OrderedSet(set):

    cdef list _list

    @classmethod
    def __class_getitem__(cls, key):
        return cls

    def __init__(self, d=None):
        set.__init__(self)
        if d is not None:
            self._list = cunique_list(d)
            set.update(self, self._list)
        else:
            self._list = []

    cpdef OrderedSet copy(self):
        cdef OrderedSet cp = OrderedSet.__new__(OrderedSet)
        cp._list = list(self._list)
        set.update(cp, cp._list)
        return cp

    @cython.final
    cdef OrderedSet _from_list(self, list new_list):
        cdef OrderedSet new = OrderedSet.__new__(OrderedSet)
        new._list = new_list
        set.update(new, new_list)
        return new

    def add(self, element):
        if element not in self:
            self._list.append(element)
            PySet_Add(self, element)

    def remove(self, element):
        # set.remove will raise if element is not in self
        set.remove(self, element)
        self._list.remove(element)

    def pop(self):
        try:
            value = self._list.pop()
        except IndexError:
            raise KeyError("pop from an empty set") from None
        set.remove(self, value)
        return value

    def insert(self, Py_ssize_t pos, element):
        if element not in self:
            self._list.insert(pos, element)
            PySet_Add(self, element)

    def discard(self, element):
        if element in self:
            set.remove(self, element)
            self._list.remove(element)

    def clear(self):
        set.clear(self)
        self._list = []

    def __getitem__(self, key):
        return self._list[key]

    def __iter__(self):
        return iter(self._list)

    def __add__(self, other):
        return self.union(other)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._list)

    __str__ = __repr__

    def update(self, *iterables):
        for iterable in iterables:
            for e in iterable:
                if e not in self:
                    self._list.append(e)
                    set.add(self, e)

    def __ior__(self, iterable):
        self.update(iterable)
        return self

    def union(self, *other):
        result = self.copy()
        result.update(*other)
        return result

    def __or__(self, other):
        return self.union(other)

    def intersection(self, *other):
        cdef set other_set = set.intersection(self, *other)
        return self._from_list([a for a in self._list if a in other_set])

    def __and__(self, other):
        return self.intersection(other)

    def symmetric_difference(self, other):
        cdef set other_set
        if isinstance(other, set):
            other_set = <set> other
            collection = other_set
        elif isinstance(other, Collection):
            collection = other
            other_set = set(other)
        else:
            collection = list(other)
            other_set = set(collection)
        result = self._from_list([a for a in self._list if a not in other_set])
        result.update(a for a in collection if a not in self)
        return result

    def __xor__(self, other):
        return self.symmetric_difference(other)

    def difference(self, *other):
        cdef set other_set = set.difference(self, *other)
        return self._from_list([a for a in self._list if a in other_set])

    def __sub__(self, other):
        return self.difference(other)

    def intersection_update(self, *other):
        set.intersection_update(self, *other)
        self._list = [a for a in self._list if a in self]

    def __iand__(self, other):
        self.intersection_update(other)
        return self

    cpdef symmetric_difference_update(self, other):
        collection = other if isinstance(other, Collection) else list(other)
        set.symmetric_difference_update(self, collection)
        self._list = [a for a in self._list if a in self]
        self._list += [a for a in collection if a in self]

    def __ixor__(self, other):
        self.symmetric_difference_update(other)
        return self

    def difference_update(self, *other):
        set.difference_update(self, *other)
        self._list = [a for a in self._list if a in self]

    def __isub__(self, other):
        self.difference_update(other)
        return self

cdef object cy_id(object item):
    return PyLong_FromLongLong(<long long> (<void *>item))

# NOTE: cython 0.x will call __add__, __sub__, etc with the parameter swapped
# instead of the __rmeth__, so they need to check that also self is of the
# correct type. This is fixed in cython 3.x. See:
# https://docs.cython.org/en/latest/src/userguide/special_methods.html#arithmetic-methods
cdef class IdentitySet:
    """A set that considers only object id() for uniqueness.

    This strategy has edge cases for builtin types- it's possible to have
    two 'foo' strings in one of these sets, for example.  Use sparingly.

    """

    cdef dict _members

    def __init__(self, iterable=None):
        self._members = {}
        if iterable:
            self.update(iterable)

    def add(self, value):
        self._members[cy_id(value)] = value

    def __contains__(self, value):
        return cy_id(value) in self._members

    cpdef remove(self, value):
        del self._members[cy_id(value)]

    def discard(self, value):
        try:
            self.remove(value)
        except KeyError:
            pass

    def pop(self):
        cdef tuple pair
        try:
            pair = self._members.popitem()
            return pair[1]
        except KeyError:
            raise KeyError("pop from an empty set")

    def clear(self):
        self._members.clear()

    def __eq__(self, other):
        cdef IdentitySet other_
        if isinstance(other, IdentitySet):
            other_ = other
            return self._members == other_._members
        else:
            return False

    def __ne__(self, other):
        cdef IdentitySet other_
        if isinstance(other, IdentitySet):
            other_ = other
            return self._members != other_._members
        else:
            return True

    cpdef issubset(self, iterable):
        cdef IdentitySet other
        if isinstance(iterable, self.__class__):
            other = iterable
        else:
            other = self.__class__(iterable)

        if len(self) > len(other):
            return False
        for m in filterfalse(other._members.__contains__, self._members):
            return False
        return True

    def __le__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        return self.issubset(other)

    def __lt__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        return len(self) < len(other) and self.issubset(other)

    cpdef issuperset(self, iterable):
        cdef IdentitySet other
        if isinstance(iterable, self.__class__):
            other = iterable
        else:
            other = self.__class__(iterable)

        if len(self) < len(other):
            return False
        for m in filterfalse(self._members.__contains__, other._members):
            return False
        return True

    def __ge__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        return self.issuperset(other)

    def __gt__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        return len(self) > len(other) and self.issuperset(other)

    cpdef IdentitySet union(self, iterable):
        cdef IdentitySet result = self.__class__()
        result._members.update(self._members)
        result.update(iterable)
        return result

    def __or__(self, other):
        if not isinstance(other, IdentitySet) or not isinstance(self, IdentitySet):
            return NotImplemented
        return self.union(other)

    cpdef update(self, iterable):
        for obj in iterable:
            self._members[cy_id(obj)] = obj

    def __ior__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        self.update(other)
        return self

    cpdef IdentitySet difference(self, iterable):
        cdef IdentitySet result = self.__new__(self.__class__)
        if isinstance(iterable, self.__class__):
            other = (<IdentitySet>iterable)._members
        else:
            other = {cy_id(obj) for obj in iterable}
        result._members = {k:v for k, v in self._members.items() if k not in other}
        return result

    def __sub__(self, other):
        if not isinstance(other, IdentitySet) or not isinstance(self, IdentitySet):
            return NotImplemented
        return self.difference(other)

    cpdef difference_update(self, iterable):
        cdef IdentitySet other = self.difference(iterable)
        self._members = other._members

    def __isub__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        self.difference_update(other)
        return self

    cpdef IdentitySet intersection(self, iterable):
        cdef IdentitySet result = self.__new__(self.__class__)
        if isinstance(iterable, self.__class__):
            other = (<IdentitySet>iterable)._members
        else:
            other = {cy_id(obj) for obj in iterable}
        result._members = {k: v for k, v in self._members.items() if k in other}
        return result

    def __and__(self, other):
        if not isinstance(other, IdentitySet) or not isinstance(self, IdentitySet):
            return NotImplemented
        return self.intersection(other)

    cpdef intersection_update(self, iterable):
        cdef IdentitySet other = self.intersection(iterable)
        self._members = other._members

    def __iand__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        self.intersection_update(other)
        return self

    cpdef IdentitySet symmetric_difference(self, iterable):
        cdef IdentitySet result = self.__new__(self.__class__)
        cdef dict other
        if isinstance(iterable, self.__class__):
            other = (<IdentitySet>iterable)._members
        else:
            other = {cy_id(obj): obj for obj in iterable}
        result._members = {k: v for k, v in self._members.items() if k not in other}
        result._members.update(
            [(k, v) for k, v in other.items() if k not in self._members]
        )
        return result

    def __xor__(self, other):
        if not isinstance(other, IdentitySet) or not isinstance(self, IdentitySet):
            return NotImplemented
        return self.symmetric_difference(other)

    cpdef symmetric_difference_update(self, iterable):
        cdef IdentitySet other = self.symmetric_difference(iterable)
        self._members = other._members

    def __ixor__(self, other):
        if not isinstance(other, IdentitySet):
            return NotImplemented
        self.symmetric_difference(other)
        return self

    cpdef IdentitySet copy(self):
        cdef IdentitySet cp = self.__new__(self.__class__)
        cp._members = self._members.copy()
        return cp

    def __copy__(self):
        return self.copy()

    def __len__(self):
        return len(self._members)

    def __iter__(self):
        return iter(self._members.values())

    def __hash__(self):
        raise TypeError("set objects are unhashable")

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, list(self._members.values()))



#!/usr/bin/env python
#
# Very simple serial terminal
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C)2002-2020 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause

from __future__ import absolute_import

import codecs
import os
import sys
import threading

import serial
from serial.tools.list_ports import comports
from serial.tools import hexlify_codec

# pylint: disable=wrong-import-order,wrong-import-position

codecs.register(lambda c: hexlify_codec.getregentry() if c == 'hexlify' else None)

try:
    raw_input
except NameError:
    # pylint: disable=redefined-builtin,invalid-name
    raw_input = input   # in python3 it's "raw"
    unichr = chr


def key_description(character):
    """generate a readable description for a key"""
    ascii_code = ord(character)
    if ascii_code < 32:
        return 'Ctrl+{:c}'.format(ord('@') + ascii_code)
    else:
        return repr(character)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class ConsoleBase(object):
    """OS abstraction for console (input/output codec, no echo)"""

    def __init__(self):
        if sys.version_info >= (3, 0):
            self.byte_output = sys.stdout.buffer
        else:
            self.byte_output = sys.stdout
        self.output = sys.stdout

    def setup(self):
        """Set console to read single characters, no echo"""

    def cleanup(self):
        """Restore default console settings"""

    def getkey(self):
        """Read a single key from the console"""
        return None

    def write_bytes(self, byte_string):
        """Write bytes (already encoded)"""
        self.byte_output.write(byte_string)
        self.byte_output.flush()

    def write(self, text):
        """Write string"""
        self.output.write(text)
        self.output.flush()

    def cancel(self):
        """Cancel getkey operation"""

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    # context manager:
    # switch terminal temporary to normal mode (e.g. to get user input)

    def __enter__(self):
        self.cleanup()
        return self

    def __exit__(self, *args, **kwargs):
        self.setup()


if os.name == 'nt':  # noqa
    import msvcrt
    import ctypes
    import platform

    class Out(object):
        """file-like wrapper that uses os.write"""

        def __init__(self, fd):
            self.fd = fd

        def flush(self):
            pass

        def write(self, s):
            os.write(self.fd, s)

    class Console(ConsoleBase):
        fncodes = {
            ';': '\1bOP',  # F1
            '<': '\1bOQ',  # F2
            '=': '\1bOR',  # F3
            '>': '\1bOS',  # F4
            '?': '\1b[15~',  # F5
            '@': '\1b[17~',  # F6
            'A': '\1b[18~',  # F7
            'B': '\1b[19~',  # F8
            'C': '\1b[20~',  # F9
            'D': '\1b[21~',  # F10
        }
        navcodes = {
            'H': '\x1b[A',  # UP
            'P': '\x1b[B',  # DOWN
            'K': '\x1b[D',  # LEFT
            'M': '\x1b[C',  # RIGHT
            'G': '\x1b[H',  # HOME
            'O': '\x1b[F',  # END
            'R': '\x1b[2~',  # INSERT
            'S': '\x1b[3~',  # DELETE
            'I': '\x1b[5~',  # PGUP
            'Q': '\x1b[6~',  # PGDN        
        }
        
        def __init__(self):
            super(Console, self).__init__()
            self._saved_ocp = ctypes.windll.kernel32.GetConsoleOutputCP()
            self._saved_icp = ctypes.windll.kernel32.GetConsoleCP()
            ctypes.windll.kernel32.SetConsoleOutputCP(65001)
            ctypes.windll.kernel32.SetConsoleCP(65001)
            # ANSI handling available through SetConsoleMode since Windows 10 v1511 
            # https://en.wikipedia.org/wiki/ANSI_escape_code#cite_note-win10th2-1
            if platform.release() == '10' and int(platform.version().split('.')[2]) > 10586:
                ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
                import ctypes.wintypes as wintypes
                if not hasattr(wintypes, 'LPDWORD'): # PY2
                    wintypes.LPDWORD = ctypes.POINTER(wintypes.DWORD)
                SetConsoleMode = ctypes.windll.kernel32.SetConsoleMode
                GetConsoleMode = ctypes.windll.kernel32.GetConsoleMode
                GetStdHandle = ctypes.windll.kernel32.GetStdHandle
                mode = wintypes.DWORD()
                GetConsoleMode(GetStdHandle(-11), ctypes.byref(mode))
                if (mode.value & ENABLE_VIRTUAL_TERMINAL_PROCESSING) == 0:
                    SetConsoleMode(GetStdHandle(-11), mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING)
                    self._saved_cm = mode
            self.output = codecs.getwriter('UTF-8')(Out(sys.stdout.fileno()), 'replace')
            # the change of the code page is not propagated to Python, manually fix it
            sys.stderr = codecs.getwriter('UTF-8')(Out(sys.stderr.fileno()), 'replace')
            sys.stdout = self.output
            self.output.encoding = 'UTF-8'  # needed for input

        def __del__(self):
            ctypes.windll.kernel32.SetConsoleOutputCP(self._saved_ocp)
            ctypes.windll.kernel32.SetConsoleCP(self._saved_icp)
            try:
                ctypes.windll.kernel32.SetConsoleMode(ctypes.windll.kernel32.GetStdHandle(-11), self._saved_cm)
            except AttributeError: # in case no _saved_cm
                pass

        def getkey(self):
            while True:
                z = msvcrt.getwch()
                if z == unichr(13):
                    return unichr(10)
                elif z is unichr(0) or z is unichr(0xe0):
                    try:
                        code = msvcrt.getwch()
                        if z is unichr(0):
                            return self.fncodes[code]
                        else:
                            return self.navcodes[code]
                    except KeyError:
                        pass
                else:
                    return z

        def cancel(self):
            # CancelIo, CancelSynchronousIo do not seem to work when using
            # getwch, so instead, send a key to the window with the console
            hwnd = ctypes.windll.kernel32.GetConsoleWindow()
            ctypes.windll.user32.PostMessageA(hwnd, 0x100, 0x0d, 0)

elif os.name == 'posix':
    import atexit
    import termios
    import fcntl

    class Console(ConsoleBase):
        def __init__(self):
            super(Console, self).__init__()
            self.fd = sys.stdin.fileno()
            self.old = termios.tcgetattr(self.fd)
            atexit.register(self.cleanup)
            if sys.version_info < (3, 0):
                self.enc_stdin = codecs.getreader(sys.stdin.encoding)(sys.stdin)
            else:
                self.enc_stdin = sys.stdin

        def setup(self):
            new = termios.tcgetattr(self.fd)
            new[3] = new[3] & ~termios.ICANON & ~termios.ECHO & ~termios.ISIG
            new[6][termios.VMIN] = 1
            new[6][termios.VTIME] = 0
            termios.tcsetattr(self.fd, termios.TCSANOW, new)

        def getkey(self):
            c = self.enc_stdin.read(1)
            if c == unichr(0x7f):
                c = unichr(8)    # map the BS key (which yields DEL) to backspace
            return c

        def cancel(self):
            fcntl.ioctl(self.fd, termios.TIOCSTI, b'\0')

        def cleanup(self):
            termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.old)

else:
    raise NotImplementedError(
        'Sorry no implementation for your platform ({}) available.'.format(sys.platform))


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class Transform(object):
    """do-nothing: forward all data unchanged"""
    def rx(self, text):
        """text received from serial port"""
        return text

    def tx(self, text):
        """text to be sent to serial port"""
        return text

    def echo(self, text):
        """text to be sent but displayed on console"""
        return text


class CRLF(Transform):
    """ENTER sends CR+LF"""

    def tx(self, text):
        return text.replace('\n', '\r\n')


class CR(Transform):
    """ENTER sends CR"""

    def rx(self, text):
        return text.replace('\r', '\n')

    def tx(self, text):
        return text.replace('\n', '\r')


class LF(Transform):
    """ENTER sends LF"""


class NoTerminal(Transform):
    """remove typical terminal control codes from input"""

    REPLACEMENT_MAP = dict((x, 0x2400 + x) for x in range(32) if unichr(x) not in '\r\n\b\t')
    REPLACEMENT_MAP.update(
        {
            0x7F: 0x2421,  # DEL
            0x9B: 0x2425,  # CSI
        })

    def rx(self, text):
        return text.translate(self.REPLACEMENT_MAP)

    echo = rx


class NoControls(NoTerminal):
    """Remove all control codes, incl. CR+LF"""

    REPLACEMENT_MAP = dict((x, 0x2400 + x) for x in range(32))
    REPLACEMENT_MAP.update(
        {
            0x20: 0x2423,  # visual space
            0x7F: 0x2421,  # DEL
            0x9B: 0x2425,  # CSI
        })


class Printable(Transform):
    """Show decimal code for all non-ASCII characters and replace most control codes"""

    def rx(self, text):
        r = []
        for c in text:
            if ' ' <= c < '\x7f' or c in '\r\n\b\t':
                r.append(c)
            elif c < ' ':
                r.append(unichr(0x2400 + ord(c)))
            else:
                r.extend(unichr(0x2080 + ord(d) - 48) for d in '{:d}'.format(ord(c)))
                r.append(' ')
        return ''.join(r)

    echo = rx


class Colorize(Transform):
    """Apply different colors for received and echo"""

    def __init__(self):
        # XXX make it configurable, use colorama?
        self.input_color = '\x1b[37m'
        self.echo_color = '\x1b[31m'

    def rx(self, text):
        return self.input_color + text

    def echo(self, text):
        return self.echo_color + text


class DebugIO(Transform):
    """Print what is sent and received"""

    def rx(self, text):
        sys.stderr.write(' [RX:{!r}] '.format(text))
        sys.stderr.flush()
        return text

    def tx(self, text):
        sys.stderr.write(' [TX:{!r}] '.format(text))
        sys.stderr.flush()
        return text


# other ideas:
# - add date/time for each newline
# - insert newline after: a) timeout b) packet end character

EOL_TRANSFORMATIONS = {
    'crlf': CRLF,
    'cr': CR,
    'lf': LF,
}

TRANSFORMATIONS = {
    'direct': Transform,    # no transformation
    'default': NoTerminal,
    'nocontrol': NoControls,
    'printable': Printable,
    'colorize': Colorize,
    'debug': DebugIO,
}


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def ask_for_port():
    """\
    Show a list of ports and ask the user for a choice. To make selection
    easier on systems with long device names, also allow the input of an
    index.
    """
    sys.stderr.write('\n--- Available ports:\n')
    ports = []
    for n, (port, desc, hwid) in enumerate(sorted(comports()), 1):
        sys.stderr.write('--- {:2}: {:20} {!r}\n'.format(n, port, desc))
        ports.append(port)
    while True:
        port = raw_input('--- Enter port index or full name: ')
        try:
            index = int(port) - 1
            if not 0 <= index < len(ports):
                sys.stderr.write('--- Invalid index!\n')
                continue
        except ValueError:
            pass
        else:
            port = ports[index]
        return port


class Miniterm(object):
    """\
    Terminal application. Copy data from serial port to console and vice versa.
    Handle special keys from the console to show menu etc.
    """

    def __init__(self, serial_instance, echo=False, eol='crlf', filters=()):
        self.console = Console()
        self.serial = serial_instance
        self.echo = echo
        self.raw = False
        self.input_encoding = 'UTF-8'
        self.output_encoding = 'UTF-8'
        self.eol = eol
        self.filters = filters
        self.update_transformations()
        self.exit_character = unichr(0x1d)  # GS/CTRL+]
        self.menu_character = unichr(0x14)  # Menu: CTRL+T
        self.alive = None
        self._reader_alive = None
        self.receiver_thread = None
        self.rx_decoder = None
        self.tx_decoder = None

    def _start_reader(self):
        """Start reader thread"""
        self._reader_alive = True
        # start serial->console thread
        self.receiver_thread = threading.Thread(target=self.reader, name='rx')
        self.receiver_thread.daemon = True
        self.receiver_thread.start()

    def _stop_reader(self):
        """Stop reader thread only, wait for clean exit of thread"""
        self._reader_alive = False
        if hasattr(self.serial, 'cancel_read'):
            self.serial.cancel_read()
        self.receiver_thread.join()

    def start(self):
        """start worker threads"""
        self.alive = True
        self._start_reader()
        # enter console->serial loop
        self.transmitter_thread = threading.Thread(target=self.writer, name='tx')
        self.transmitter_thread.daemon = True
        self.transmitter_thread.start()
        self.console.setup()

    def stop(self):
        """set flag to stop worker threads"""
        self.alive = False

    def join(self, transmit_only=False):
        """wait for worker threads to terminate"""
        self.transmitter_thread.join()
        if not transmit_only:
            if hasattr(self.serial, 'cancel_read'):
                self.serial.cancel_read()
            self.receiver_thread.join()

    def close(self):
        self.serial.close()

    def update_transformations(self):
        """take list of transformation classes and instantiate them for rx and tx"""
        transformations = [EOL_TRANSFORMATIONS[self.eol]] + [TRANSFORMATIONS[f]
                                                             for f in self.filters]
        self.tx_transformations = [t() for t in transformations]
        self.rx_transformations = list(reversed(self.tx_transformations))

    def set_rx_encoding(self, encoding, errors='replace'):
        """set encoding for received data"""
        self.input_encoding = encoding
        self.rx_decoder = codecs.getincrementaldecoder(encoding)(errors)

    def set_tx_encoding(self, encoding, errors='replace'):
        """set encoding for transmitted data"""
        self.output_encoding = encoding
        self.tx_encoder = codecs.getincrementalencoder(encoding)(errors)

    def dump_port_settings(self):
        """Write current settings to sys.stderr"""
        sys.stderr.write("\n--- Settings: {p.name}  {p.baudrate},{p.bytesize},{p.parity},{p.stopbits}\n".format(
            p=self.serial))
        sys.stderr.write('--- RTS: {:8}  DTR: {:8}  BREAK: {:8}\n'.format(
            ('active' if self.serial.rts else 'inactive'),
            ('active' if self.serial.dtr else 'inactive'),
            ('active' if self.serial.break_condition else 'inactive')))
        try:
            sys.stderr.write('--- CTS: {:8}  DSR: {:8}  RI: {:8}  CD: {:8}\n'.format(
                ('active' if self.serial.cts else 'inactive'),
                ('active' if self.serial.dsr else 'inactive'),
                ('active' if self.serial.ri else 'inactive'),
                ('active' if self.serial.cd else 'inactive')))
        except serial.SerialException:
            # on RFC 2217 ports, it can happen if no modem state notification was
            # yet received. ignore this error.
            pass
        sys.stderr.write('--- software flow control: {}\n'.format('active' if self.serial.xonxoff else 'inactive'))
        sys.stderr.write('--- hardware flow control: {}\n'.format('active' if self.serial.rtscts else 'inactive'))
        sys.stderr.write('--- serial input encoding: {}\n'.format(self.input_encoding))
        sys.stderr.write('--- serial output encoding: {}\n'.format(self.output_encoding))
        sys.stderr.write('--- EOL: {}\n'.format(self.eol.upper()))
        sys.stderr.write('--- filters: {}\n'.format(' '.join(self.filters)))

    def reader(self):
        """loop and copy serial->console"""
        try:
            while self.alive and self._reader_alive:
                # read all that is there or wait for one byte
                data = self.serial.read(self.serial.in_waiting or 1)
                if data:
                    if self.raw:
                        self.console.write_bytes(data)
                    else:
                        text = self.rx_decoder.decode(data)
                        for transformation in self.rx_transformations:
                            text = transformation.rx(text)
                        self.console.write(text)
        except serial.SerialException:
            self.alive = False
            self.console.cancel()
            raise       # XXX handle instead of re-raise?

    def writer(self):
        """\
        Loop and copy console->serial until self.exit_character character is
        found. When self.menu_character is found, interpret the next key
        locally.
        """
        menu_active = False
        try:
            while self.alive:
                try:
                    c = self.console.getkey()
                except KeyboardInterrupt:
                    c = '\x03'
                if not self.alive:
                    break
                if menu_active:
                    self.handle_menu_key(c)
                    menu_active = False
                elif c == self.menu_character:
                    menu_active = True      # next char will be for menu
                elif c == self.exit_character:
                    self.stop()             # exit app
                    break
                else:
                    #~ if self.raw:
                    text = c
                    for transformation in self.tx_transformations:
                        text = transformation.tx(text)
                    self.serial.write(self.tx_encoder.encode(text))
                    if self.echo:
                        echo_text = c
                        for transformation in self.tx_transformations:
                            echo_text = transformation.echo(echo_text)
                        self.console.write(echo_text)
        except:
            self.alive = False
            raise

    def handle_menu_key(self, c):
        """Implement a simple menu / settings"""
        if c == self.menu_character or c == self.exit_character:
            # Menu/exit character again -> send itself
            self.serial.write(self.tx_encoder.encode(c))
            if self.echo:
                self.console.write(c)
        elif c == '\x15':                       # CTRL+U -> upload file
            self.upload_file()
        elif c in '\x08hH?':                    # CTRL+H, h, H, ? -> Show help
            sys.stderr.write(self.get_help_text())
        elif c == '\x12':                       # CTRL+R -> Toggle RTS
            self.serial.rts = not self.serial.rts
            sys.stderr.write('--- RTS {} ---\n'.format('active' if self.serial.rts else 'inactive'))
        elif c == '\x04':                       # CTRL+D -> Toggle DTR
            self.serial.dtr = not self.serial.dtr
            sys.stderr.write('--- DTR {} ---\n'.format('active' if self.serial.dtr else 'inactive'))
        elif c == '\x02':                       # CTRL+B -> toggle BREAK condition
            self.serial.break_condition = not self.serial.break_condition
            sys.stderr.write('--- BREAK {} ---\n'.format('active' if self.serial.break_condition else 'inactive'))
        elif c == '\x05':                       # CTRL+E -> toggle local echo
            self.echo = not self.echo
            sys.stderr.write('--- local echo {} ---\n'.format('active' if self.echo else 'inactive'))
        elif c == '\x06':                       # CTRL+F -> edit filters
            self.change_filter()
        elif c == '\x0c':                       # CTRL+L -> EOL mode
            modes = list(EOL_TRANSFORMATIONS)   # keys
            eol = modes.index(self.eol) + 1
            if eol >= len(modes):
                eol = 0
            self.eol = modes[eol]
            sys.stderr.write('--- EOL: {} ---\n'.format(self.eol.upper()))
            self.update_transformations()
        elif c == '\x01':                       # CTRL+A -> set encoding
            self.change_encoding()
        elif c == '\x09':                       # CTRL+I -> info
            self.dump_port_settings()
        #~ elif c == '\x01':                       # CTRL+A -> cycle escape mode
        #~ elif c == '\x0c':                       # CTRL+L -> cycle linefeed mode
        elif c in 'pP':                         # P -> change port
            self.change_port()
        elif c in 'zZ':                         # S -> suspend / open port temporarily
            self.suspend_port()
        elif c in 'bB':                         # B -> change baudrate
            self.change_baudrate()
        elif c == '8':                          # 8 -> change to 8 bits
            self.serial.bytesize = serial.EIGHTBITS
            self.dump_port_settings()
        elif c == '7':                          # 7 -> change to 8 bits
            self.serial.bytesize = serial.SEVENBITS
            self.dump_port_settings()
        elif c in 'eE':                         # E -> change to even parity
            self.serial.parity = serial.PARITY_EVEN
            self.dump_port_settings()
        elif c in 'oO':                         # O -> change to odd parity
            self.serial.parity = serial.PARITY_ODD
            self.dump_port_settings()
        elif c in 'mM':                         # M -> change to mark parity
            self.serial.parity = serial.PARITY_MARK
            self.dump_port_settings()
        elif c in 'sS':                         # S -> change to space parity
            self.serial.parity = serial.PARITY_SPACE
            self.dump_port_settings()
        elif c in 'nN':                         # N -> change to no parity
            self.serial.parity = serial.PARITY_NONE
            self.dump_port_settings()
        elif c == '1':                          # 1 -> change to 1 stop bits
            self.serial.stopbits = serial.STOPBITS_ONE
            self.dump_port_settings()
        elif c == '2':                          # 2 -> change to 2 stop bits
            self.serial.stopbits = serial.STOPBITS_TWO
            self.dump_port_settings()
        elif c == '3':                          # 3 -> change to 1.5 stop bits
            self.serial.stopbits = serial.STOPBITS_ONE_POINT_FIVE
            self.dump_port_settings()
        elif c in 'xX':                         # X -> change software flow control
            self.serial.xonxoff = (c == 'X')
            self.dump_port_settings()
        elif c in 'rR':                         # R -> change hardware flow control
            self.serial.rtscts = (c == 'R')
            self.dump_port_settings()
        elif c in 'qQ':
            self.stop()                         # Q -> exit app
        else:
            sys.stderr.write('--- unknown menu character {} --\n'.format(key_description(c)))

    def upload_file(self):
        """Ask user for filenname and send its contents"""
        sys.stderr.write('\n--- File to upload: ')
        sys.stderr.flush()
        with self.console:
            filename = sys.stdin.readline().rstrip('\r\n')
            if filename:
                try:
                    with open(filename, 'rb') as f:
                        sys.stderr.write('--- Sending file {} ---\n'.format(filename))
                        while True:
                            block = f.read(1024)
                            if not block:
                                break
                            self.serial.write(block)
                            # Wait for output buffer to drain.
                            self.serial.flush()
                            sys.stderr.write('.')   # Progress indicator.
                    sys.stderr.write('\n--- File {} sent ---\n'.format(filename))
                except IOError as e:
                    sys.stderr.write('--- ERROR opening file {}: {} ---\n'.format(filename, e))

    def change_filter(self):
        """change the i/o transformations"""
        sys.stderr.write('\n--- Available Filters:\n')
        sys.stderr.write('\n'.join(
            '---   {:<10} = {.__doc__}'.format(k, v)
            for k, v in sorted(TRANSFORMATIONS.items())))
        sys.stderr.write('\n--- Enter new filter name(s) [{}]: '.format(' '.join(self.filters)))
        with self.console:
            new_filters = sys.stdin.readline().lower().split()
        if new_filters:
            for f in new_filters:
                if f not in TRANSFORMATIONS:
                    sys.stderr.write('--- unknown filter: {!r}\n'.format(f))
                    break
            else:
                self.filters = new_filters
                self.update_transformations()
        sys.stderr.write('--- filters: {}\n'.format(' '.join(self.filters)))

    def change_encoding(self):
        """change encoding on the serial port"""
        sys.stderr.write('\n--- Enter new encoding name [{}]: '.format(self.input_encoding))
        with self.console:
            new_encoding = sys.stdin.readline().strip()
        if new_encoding:
            try:
                codecs.lookup(new_encoding)
            except LookupError:
                sys.stderr.write('--- invalid encoding name: {}\n'.format(new_encoding))
            else:
                self.set_rx_encoding(new_encoding)
                self.set_tx_encoding(new_encoding)
        sys.stderr.write('--- serial input encoding: {}\n'.format(self.input_encoding))
        sys.stderr.write('--- serial output encoding: {}\n'.format(self.output_encoding))

    def change_baudrate(self):
        """change the baudrate"""
        sys.stderr.write('\n--- Baudrate: ')
        sys.stderr.flush()
        with self.console:
            backup = self.serial.baudrate
            try:
                self.serial.baudrate = int(sys.stdin.readline().strip())
            except ValueError as e:
                sys.stderr.write('--- ERROR setting baudrate: {} ---\n'.format(e))
                self.serial.baudrate = backup
            else:
                self.dump_port_settings()

    def change_port(self):
        """Have a conversation with the user to change the serial port"""
        with self.console:
            try:
                port = ask_for_port()
            except KeyboardInterrupt:
                port = None
        if port and port != self.serial.port:
            # reader thread needs to be shut down
            self._stop_reader()
            # save settings
            settings = self.serial.getSettingsDict()
            try:
                new_serial = serial.serial_for_url(port, do_not_open=True)
                # restore settings and open
                new_serial.applySettingsDict(settings)
                new_serial.rts = self.serial.rts
                new_serial.dtr = self.serial.dtr
                new_serial.open()
                new_serial.break_condition = self.serial.break_condition
            except Exception as e:
                sys.stderr.write('--- ERROR opening new port: {} ---\n'.format(e))
                new_serial.close()
            else:
                self.serial.close()
                self.serial = new_serial
                sys.stderr.write('--- Port changed to: {} ---\n'.format(self.serial.port))
            # and restart the reader thread
            self._start_reader()

    def suspend_port(self):
        """\
        open port temporarily, allow reconnect, exit and port change to get
        out of the loop
        """
        # reader thread needs to be shut down
        self._stop_reader()
        self.serial.close()
        sys.stderr.write('\n--- Port closed: {} ---\n'.format(self.serial.port))
        do_change_port = False
        while not self.serial.is_open:
            sys.stderr.write('--- Quit: {exit} | p: port change | any other key to reconnect ---\n'.format(
                exit=key_description(self.exit_character)))
            k = self.console.getkey()
            if k == self.exit_character:
                self.stop()             # exit app
                break
            elif k in 'pP':
                do_change_port = True
                break
            try:
                self.serial.open()
            except Exception as e:
                sys.stderr.write('--- ERROR opening port: {} ---\n'.format(e))
        if do_change_port:
            self.change_port()
        else:
            # and restart the reader thread
            self._start_reader()
            sys.stderr.write('--- Port opened: {} ---\n'.format(self.serial.port))

    def get_help_text(self):
        """return the help text"""
        # help text, starts with blank line!
        return """
--- pySerial ({version}) - miniterm - help
---
--- {exit:8} Exit program (alias {menu} Q)
--- {menu:8} Menu escape key, followed by:
--- Menu keys:
---    {menu:7} Send the menu character itself to remote
---    {exit:7} Send the exit character itself to remote
---    {info:7} Show info
---    {upload:7} Upload file (prompt will be shown)
---    {repr:7} encoding
---    {filter:7} edit filters
--- Toggles:
---    {rts:7} RTS   {dtr:7} DTR   {brk:7} BREAK
---    {echo:7} echo  {eol:7} EOL
---
--- Port settings ({menu} followed by the following):
---    p          change port
---    7 8        set data bits
---    N E O S M  change parity (None, Even, Odd, Space, Mark)
---    1 2 3      set stop bits (1, 2, 1.5)
---    b          change baud rate
---    x X        disable/enable software flow control
---    r R        disable/enable hardware flow control
""".format(version=getattr(serial, 'VERSION', 'unknown version'),
           exit=key_description(self.exit_character),
           menu=key_description(self.menu_character),
           rts=key_description('\x12'),
           dtr=key_description('\x04'),
           brk=key_description('\x02'),
           echo=key_description('\x05'),
           info=key_description('\x09'),
           upload=key_description('\x15'),
           repr=key_description('\x01'),
           filter=key_description('\x06'),
           eol=key_description('\x0c'))


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# default args can be used to override when calling main() from an other script
# e.g to create a miniterm-my-device.py
def main(default_port=None, default_baudrate=9600, default_rts=None, default_dtr=None):
    """Command line tool, entry point"""

    import argparse

    parser = argparse.ArgumentParser(
        description='Miniterm - A simple terminal program for the serial port.')

    parser.add_argument(
        'port',
        nargs='?',
        help='serial port name ("-" to show port list)',
        default=default_port)

    parser.add_argument(
        'baudrate',
        nargs='?',
        type=int,
        help='set baud rate, default: %(default)s',
        default=default_baudrate)

    group = parser.add_argument_group('port settings')

    group.add_argument(
        '--parity',
        choices=['N', 'E', 'O', 'S', 'M'],
        type=lambda c: c.upper(),
        help='set parity, one of {N E O S M}, default: N',
        default='N')

    group.add_argument(
        '--rtscts',
        action='store_true',
        help='enable RTS/CTS flow control (default off)',
        default=False)

    group.add_argument(
        '--xonxoff',
        action='store_true',
        help='enable software flow control (default off)',
        default=False)

    group.add_argument(
        '--rts',
        type=int,
        help='set initial RTS line state (possible values: 0, 1)',
        default=default_rts)

    group.add_argument(
        '--dtr',
        type=int,
        help='set initial DTR line state (possible values: 0, 1)',
        default=default_dtr)

    group.add_argument(
        '--non-exclusive',
        dest='exclusive',
        action='store_false',
        help='disable locking for native ports',
        default=True)

    group.add_argument(
        '--ask',
        action='store_true',
        help='ask again for port when open fails',
        default=False)

    group = parser.add_argument_group('data handling')

    group.add_argument(
        '-e', '--echo',
        action='store_true',
        help='enable local echo (default off)',
        default=False)

    group.add_argument(
        '--encoding',
        dest='serial_port_encoding',
        metavar='CODEC',
        help='set the encoding for the serial port (e.g. hexlify, Latin1, UTF-8), default: %(default)s',
        default='UTF-8')

    group.add_argument(
        '-f', '--filter',
        action='append',
        metavar='NAME',
        help='add text transformation',
        default=[])

    group.add_argument(
        '--eol',
        choices=['CR', 'LF', 'CRLF'],
        type=lambda c: c.upper(),
        help='end of line mode',
        default='CRLF')

    group.add_argument(
        '--raw',
        action='store_true',
        help='Do no apply any encodings/transformations',
        default=False)

    group = parser.add_argument_group('hotkeys')

    group.add_argument(
        '--exit-char',
        type=int,
        metavar='NUM',
        help='Unicode of special character that is used to exit the application, default: %(default)s',
        default=0x1d)  # GS/CTRL+]

    group.add_argument(
        '--menu-char',
        type=int,
        metavar='NUM',
        help='Unicode code of special character that is used to control miniterm (menu), default: %(default)s',
        default=0x14)  # Menu: CTRL+T

    group = parser.add_argument_group('diagnostics')

    group.add_argument(
        '-q', '--quiet',
        action='store_true',
        help='suppress non-error messages',
        default=False)

    group.add_argument(
        '--develop',
        action='store_true',
        help='show Python traceback on error',
        default=False)

    args = parser.parse_args()

    if args.menu_char == args.exit_char:
        parser.error('--exit-char can not be the same as --menu-char')

    if args.filter:
        if 'help' in args.filter:
            sys.stderr.write('Available filters:\n')
            sys.stderr.write('\n'.join(
                '{:<10} = {.__doc__}'.format(k, v)
                for k, v in sorted(TRANSFORMATIONS.items())))
            sys.stderr.write('\n')
            sys.exit(1)
        filters = args.filter
    else:
        filters = ['default']

    while True:
        # no port given on command line -> ask user now
        if args.port is None or args.port == '-':
            try:
                args.port = ask_for_port()
            except KeyboardInterrupt:
                sys.stderr.write('\n')
                parser.error('user aborted and port is not given')
            else:
                if not args.port:
                    parser.error('port is not given')
        try:
            serial_instance = serial.serial_for_url(
                args.port,
                args.baudrate,
                parity=args.parity,
                rtscts=args.rtscts,
                xonxoff=args.xonxoff,
                do_not_open=True)

            if not hasattr(serial_instance, 'cancel_read'):
                # enable timeout for alive flag polling if cancel_read is not available
                serial_instance.timeout = 1

            if args.dtr is not None:
                if not args.quiet:
                    sys.stderr.write('--- forcing DTR {}\n'.format('active' if args.dtr else 'inactive'))
                serial_instance.dtr = args.dtr
            if args.rts is not None:
                if not args.quiet:
                    sys.stderr.write('--- forcing RTS {}\n'.format('active' if args.rts else 'inactive'))
                serial_instance.rts = args.rts

            if isinstance(serial_instance, serial.Serial):
                serial_instance.exclusive = args.exclusive

            serial_instance.open()
        except serial.SerialException as e:
            sys.stderr.write('could not open port {!r}: {}\n'.format(args.port, e))
            if args.develop:
                raise
            if not args.ask:
                sys.exit(1)
            else:
                args.port = '-'
        else:
            break

    miniterm = Miniterm(
        serial_instance,
        echo=args.echo,
        eol=args.eol.lower(),
        filters=filters)
    miniterm.exit_character = unichr(args.exit_char)
    miniterm.menu_character = unichr(args.menu_char)
    miniterm.raw = args.raw
    miniterm.set_rx_encoding(args.serial_port_encoding)
    miniterm.set_tx_encoding(args.serial_port_encoding)

    if not args.quiet:
        sys.stderr.write('--- Miniterm on {p.name}  {p.baudrate},{p.bytesize},{p.parity},{p.stopbits} ---\n'.format(
            p=miniterm.serial))
        sys.stderr.write('--- Quit: {} | Menu: {} | Help: {} followed by {} ---\n'.format(
            key_description(miniterm.exit_character),
            key_description(miniterm.menu_character),
            key_description(miniterm.menu_character),
            key_description('\x08')))

    miniterm.start()
    try:
        miniterm.join(True)
    except KeyboardInterrupt:
        pass
    if not args.quiet:
        sys.stderr.write('\n--- exit ---\n')
    miniterm.join()
    miniterm.close()

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    main()



import json

from typing import Any

from application.interactor import UserUpdaterInteractor
from application.model import AsyncUuidMessageProcessor
from application.model.customexception import SomeUserLogged
from messagebus import DefaultToken, DataType, Message
from application.util import settings


class UserUpdaterProcessor(AsyncUuidMessageProcessor):
    def __init__(self, interactor):
        # type: (UserUpdaterInteractor) -> None

        super(UserUpdaterProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (Any, Any) -> Any

        logged_users = json.loads(data)
        return logged_users

    def call_business(self, logged_users):
        # type: (Any) -> Any

        settings.clear_status()
        return self.interactor.update_users(logged_users)

    def format_response(self, message_bus, message, event, _, response):
        token = DefaultToken.TK_SYS_ACK
        if response is not None and "users_disable_logged" in response:
            token = DefaultToken.TK_SYS_NAK
            response = str(response)

        reply_message = Message(token, DataType.param, response)
        if message:
            message_bus.reply_message(message, reply_message)
        elif event:
            message_bus.reply_event(event, reply_message)

    def format_exception(self, message_bus, message, event, _, exception):
        self._reply_exception(event, exception, message, message_bus)

    def format_parse_exception(self, message_bus, message, event, _, exception):
        self._reply_exception(event, exception, message, message_bus)

    @staticmethod
    def _reply_exception(event, exception, message, message_bus):
        reply_message = Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception))
        if message:
            data = str(type(exception).__name__)
            if isinstance(exception, SomeUserLogged):
                data = str(type(exception).__name__) + "|" + exception.message

            reply_message = Message(DefaultToken.TK_SYS_NAK, DataType.param, data)
            message_bus.reply_message(message, reply_message)
        elif event:
            message_bus.reply_event(event, reply_message)
